import math
from eyed3.utils import formatSize
from eyed3.utils.console import printMsg, getTtySize
from eyed3.plugins import LoaderPlugin


class LameInfoPlugin(LoaderPlugin):
    NAMES = ["lameinfo", "xing"]
    SUMMARY = "Outputs lame header (if one exists) for file."
    DESCRIPTION = (
        "The 'lame' (or xing) header provides extra information about the mp3 "
        "that is useful to players and encoders but not officially part of "
        "the mp3 specification. Variable bit rate mp3s, for example, use this "
        "header.\n\n"
        "For more details see `here <http://gabriel.mp3-tech.org/mp3infotag.html>`_"
       )

    def printHeader(self, file_path):
        w = getTtySize()[1]
        printMsg(self._getFileHeader(file_path, w))
        printMsg(self._getHardRule(w))

    def handleFile(self, f, *_, **__):
        super().handleFile(f)
        if self.audio_file is None:
            return

        self.printHeader(f)
        if (self.audio_file.info is None
                or not self.audio_file.info.lame_tag):
            printMsg("No LAME Tag")
            return

        lt = self.audio_file.info.lame_tag
        if "infotag_crc" not in lt:
            try:
                printMsg(f"Encoder Version: {lt['encoder_version']}")
            except KeyError:
                pass
            return

        values = [("Encoder Version", lt['encoder_version']),
                  ("LAME Tag Revision", lt['tag_revision']),
                  ("VBR Method", lt['vbr_method']),
                  ("Lowpass Filter", lt['lowpass_filter']),
                  ]

        if "replaygain" in lt:
            try:
                peak = lt["replaygain"]["peak_amplitude"]
                db = 20 * math.log10(peak)
                val = "%.8f (%+.1f dB)" % (peak, db)
                values.append(("Peak Amplitude", val))
            except KeyError:
                pass
            for type_ in ["radio", "audiofile"]:
                try:
                    gain = lt["replaygain"][type_]
                    name = "%s Replay Gain" % gain['name'].capitalize()
                    val = "%s dB (%s)" % (gain['adjustment'],
                                          gain['originator'])
                    values.append((name, val))
                except KeyError:
                    pass

        values.append(("Encoding Flags", " ".join((lt["encoding_flags"]))))
        if lt["nogap"]:
            values.append(("No Gap", " and ".join(lt["nogap"])))
        values.append(("ATH Type", lt["ath_type"]))
        values.append(("Bitrate (%s)" % lt["bitrate"][1], lt["bitrate"][0]))
        values.append(("Encoder Delay", "%s samples" % lt["encoder_delay"]))
        values.append(("Encoder Padding", "%s samples" % lt["encoder_padding"]))
        values.append(("Noise Shaping", lt["noise_shaping"]))
        values.append(("Stereo Mode", lt["stereo_mode"]))
        values.append(("Unwise Settings", lt["unwise_settings"]))
        values.append(("Sample Frequency", lt["sample_freq"]))
        values.append(("MP3 Gain", "%s (%+.1f dB)" % (lt["mp3_gain"],
                                                      lt["mp3_gain"] * 1.5)))
        values.append(("Preset", lt["preset"]))
        values.append(("Surround Info", lt["surround_info"]))
        values.append(("Music Length", "%s" % formatSize(lt["music_length"])))
        values.append(("Music CRC-16", "%04X" % lt["music_crc"]))
        values.append(("LAME Tag CRC-16", "%04X" % lt["infotag_crc"]))

        for v in values:
            printMsg(f"{v[0]:<20}: {v[1]}")



import sys
import codecs
import locale
from .__about__ import __version__ as version

_DEFAULT_ENCODING = "latin1"
# The local encoding, used when parsing command line options, console output,
# etc. The default is always ``latin1`` if it cannot be determined, it is NOT
# the value shown.
LOCAL_ENCODING = locale.getpreferredencoding(do_setlocale=True)
if not LOCAL_ENCODING or LOCAL_ENCODING == "ANSI_X3.4-1968":  # pragma: no cover
    LOCAL_ENCODING = _DEFAULT_ENCODING

# The local file system encoding, the default is ``latin1`` if it cannot be determined.
LOCAL_FS_ENCODING = sys.getfilesystemencoding()
if not LOCAL_FS_ENCODING:  # pragma: no cover
    LOCAL_FS_ENCODING = _DEFAULT_ENCODING


class Error(Exception):
    """Base exception type for all eyed3 errors."""
    def __init__(self, *args):
        super().__init__(*args)
        if args:
            # The base class will do exactly this if len(args) == 1,
            # but not when > 1. Note, the 2.7 base class will, 3 will not.
            # Make it so.
            self.message = args[0]


from .utils.log import log                                          # noqa: E402
from .core import load                                              # noqa: E402

del sys
del codecs
del locale

__all__ = ["log", "load", "version", "LOCAL_ENCODING", "LOCAL_FS_ENCODING",
           "Error"]



import sys
import binascii
from pathlib import Path

import eyed3.id3
import eyed3.plugins
from eyed3.utils.log import getLogger

log = getLogger(__name__)


class ExtractPlugin(eyed3.plugins.LoaderPlugin):
    NAMES = ["extract"]
    SUMMARY = "Extract tags from audio files."

    def __init__(self, arg_parser):
        super().__init__(arg_parser, cache_files=True, track_images=False)
        self.arg_group.add_argument("-o", "--output-file",
                                    help="The the tag is written to this file in native format.")
        self.arg_group.add_argument("-H", "--hex", action="store_true",
                                    help="Output hexadecimal format.")
        self.arg_group.add_argument("--strip-padding", action="store_true",
                                    help="Exclude tag padding, if any.")

    def handleFile(self, f, *args, **kwargs):
        super().handleFile(f)
        if self.audio_file is None or self.audio_file.tag is None:
            return

        tag = self.audio_file.tag
        if not isinstance(tag, eyed3.id3.Tag):
            print("Only ID3 tags can be extracted currently.", file=sys.stderr)
            return 1

        with open(tag.file_info.name, "rb") as tag_file:
            if tag.version[0] != 1:
                # info.tag_size includes padding.
                tag_data = tag_file.read(tag.file_info.tag_size)
                if self.args.strip_padding and tag.file_info.tag_padding_size:
                    # --strip-padding
                    tag_data = tag_data[:-tag.file_info.tag_padding_size]
            else:
                # ID3 v1.x
                tag_data = tag_file.read()[-128:]

        if self.args.output_file:
            # --output-file

            if Path(tag.file_info.name).resolve() == Path(self.args.output_file).resolve():
                print("Input file overwriting not allowed, choose a different -o/--output-file",
                      file=sys.stderr)
                return 1

            with open(self.args.output_file, "wb") as out_file:
                out_file.write(tag_data)
        else:
            if self.args.hex:
                # --hex
                tag_data = str(binascii.hexlify(tag_data), "ascii")

            print(tag_data)



import io
import os
import hashlib
from pathlib import Path

from eyed3.utils import art
from eyed3 import log
from eyed3.mimetype import guessMimetype
from eyed3.plugins import LoaderPlugin
from eyed3.core import VARIOUS_ARTISTS
from eyed3.id3.frames import ImageFrame
from eyed3.utils import makeUniqueFileName
from eyed3.utils.console import printMsg, printWarning, cformat, Fore

DESCR_FNAME_PREFIX = "filename: "
md5_file_cache = {}


def _importMessage(missing):
    return f"Missing dependencies {missing}.  Install with `pip install eyeD3[art-plugin]`"


try:
    import PIL                                                            # noqa
    import requests
    from eyed3.plugins.lastfm import getAlbumArt
    _PLUGIN_ACTIVE = True
    _IMPORT_ERROR = None
except ImportError as ex:
    _PLUGIN_ACTIVE = False
    _IMPORT_ERROR = ex


class ArtFile(object):
    def __init__(self, file_path):
        self.art_type = art.matchArtFile(file_path)
        self.file_path = file_path
        self.id3_art_type = (art.TO_ID3_ART_TYPES[self.art_type][0]
                             if self.art_type else None)
        self._img_data = None
        self._mime_type = None

    @property
    def image_data(self):
        if self._img_data:
            return self._img_data
        with open(self.file_path, "rb") as f:
            self._img_data = f.read()
        return self._img_data

    @property
    def mime_type(self):
        if self._mime_type:
            return self._mime_type
        self._mime_type = guessMimetype(self.file_path)
        return self._mime_type


class ArtPlugin(LoaderPlugin):
    SUMMARY = "Art for albums, artists, etc."
    DESCRIPTION = ""
    NAMES = ["art"]

    def __init__(self, arg_parser):
        super(ArtPlugin, self).__init__(arg_parser, cache_files=True,
                                        track_images=True)
        self._retval = 0

        g = self.arg_group
        g.add_argument("-F", "--update-files", action="store_true",
                       help="Write art files from tag images.")
        g.add_argument("-T", "--update-tags", action="store_true",
                       help="Write tag image from art files.")
        dl_help = "Attempt to download album art if missing."
        g.add_argument("-D", "--download", action="store_true", help=dl_help)
        g.add_argument("-v", "--verbose", action="store_true",
                       help="Show detailed information for all art found.")

    def start(self, args, config):
        if not _PLUGIN_ACTIVE:
            err_msg = _importMessage([_IMPORT_ERROR.name])
            log.critical(err_msg)
            raise RuntimeError(err_msg)
        if args.update_files and args.update_tags:
            # Not using add_mutually_exclusive_group from argparse because
            # the options belong to the plugin opts group (self.arg_group)
            raise StopIteration("The --update-tags and --update-files options "
                                "are mutually exclusive, use only one at a "
                                "time.")
        super(ArtPlugin, self).start(args, config)

    def _verbose(self, s):
        if self.args.verbose:
            printMsg(s)

    def handleDirectory(self, d, _):
        global md5_file_cache
        md5_file_cache.clear()

        if not self._file_cache:
            log.debug(f"{d}: nothing to do.")
            return

        try:
            all_tags = sorted([f.tag for f in self._file_cache if f.tag],
                              key=lambda x: x.file_info.name)

            # If not deemed an album, move on.
            if len(set([t.album for t in all_tags])) > 1:
                log.debug(f"Skipping directory '{d}', non-album.")
                return

            printMsg(cformat("\nChecking: ", Fore.BLUE) + d)

            # File images
            dir_art = []
            for img_file in self._dir_images:
                img_base = os.path.basename(img_file)
                art_file = ArtFile(img_file)
                try:
                    pil_img = pilImage(img_file)
                except IOError as ex:
                    printWarning(str(ex))
                    continue

                if art_file.art_type:
                    self._verbose(
                        f"file {img_base}: {art_file.art_type}\n\t{pilImageDetails(pil_img)}")
                    dir_art.append(art_file)
                else:
                    self._verbose(f"file {img_base}: unknown (ignored)")

            if not dir_art:
                print(cformat("NONE", Fore.RED))
                self._retval += 1
            else:
                print(cformat("OK", Fore.GREEN))

            # --download handling
            if not dir_art and self.args.download:
                tag = all_tags[0]
                artists = set([t.artist for t in all_tags])
                if len(artists) > 1:
                    artist_query = VARIOUS_ARTISTS
                else:
                    artist_query = tag.album_artist or tag.artist

                try:
                    url = getAlbumArt(artist_query, tag.album)
                    print("Downloading album art...")
                    resp = requests.get(url)
                    if resp.status_code != 200:
                        raise ValueError()
                except ValueError:
                    print("Album art download not found")
                else:
                    img = pilImage(io.BytesIO(resp.content))
                    cover = Path(d) / "cover.{}".format(img.format.lower())
                    assert not cover.exists()
                    img.save(str(cover))
                    print("Save {cover}".format(cover=cover))

            # Tag images
            for tag in all_tags:
                file_base = os.path.basename(tag.file_info.name)
                for img in tag.images:
                    try:
                        pil_img = pilImage(img)
                        pil_img_details = pilImageDetails(pil_img)
                    except (OSError, IOError) as ex:
                        printWarning(str(ex))
                        continue

                    if img.picture_type in art.FROM_ID3_ART_TYPES:
                        img_type = art.FROM_ID3_ART_TYPES[img.picture_type]
                        self._verbose("tag %s: %s (Description: %s)\n\t%s" %
                                      (file_base, img_type, img.description,
                                       pil_img_details))
                        if self.args.update_files:
                            assert(not self.args.update_tags)
                            path = os.path.dirname(tag.file_info.name)
                            if img.description.startswith(DESCR_FNAME_PREFIX):
                                # Use filename from Image description
                                fname = img.description[
                                          len(DESCR_FNAME_PREFIX):].strip()
                                fname = os.path.splitext(fname)[0]
                            else:
                                fname = art.FILENAMES[img_type][0].strip("*")
                            fname = img.makeFileName(name=fname)

                            if (md5File(os.path.join(path, fname)) ==
                                    md5Data(img.image_data)):
                                printMsg("Skipping writing of %s, file "
                                         "exists and is exactly the same." %
                                         fname)
                            else:
                                img_file = makeUniqueFileName(
                                    os.path.join(path, fname),
                                    uniq=img.description)
                                printWarning("Writing %s..." % img_file)
                                with open(img_file, "wb") as fp:
                                    fp.write(img.image_data)
                    else:
                        self._verbose(
                            "tag %s: unhandled image type %d (ignored)" %
                            (file_base, img.picture_type)
                        )

            # Copy file art to tags.
            if self.args.update_tags:
                assert(not self.args.update_files)
                for tag in all_tags:
                    for art_file in dir_art:
                        art_path = os.path.basename(art_file.file_path)
                        printMsg("Copying %s to tag '%s' image" %
                                 (art_path, art_file.id3_art_type))

                        descr = "filename: %s" % os.path.splitext(art_path)[0]
                        tag.images.set(art_file.id3_art_type,
                                       art_file.image_data, art_file.mime_type,
                                       description=descr)
                    tag.save()

        finally:
            # Cleans up...
            super(ArtPlugin, self).handleDirectory(d, _)

    def handleDone(self):
        return self._retval


def pilImage(source):
    from PIL import Image

    if isinstance(source, ImageFrame):
        return Image.open(io.BytesIO(source.image_data))
    else:
        return Image.open(source)


def pilImageDetails(img):
    return "[%dx%d %s md5:%s]" % (img.size[0], img.size[1],
                                  img.format.lower(),
                                  md5Data(img.tobytes())) if img else ""


def md5Data(data):
    md5 = hashlib.md5()
    md5.update(data)
    return md5.hexdigest()


def md5File(file_name):
    """Compute md5 hash for contents of ``file_name``."""

    global md5_file_cache
    if file_name in md5_file_cache:
        return md5_file_cache[file_name]

    md5 = hashlib.md5()
    try:
        with open(file_name, "rb") as f:
            md5.update(f.read())

        md5_file_cache[file_name] = md5.hexdigest()
        return md5_file_cache[file_name]
    except IOError:
        return None



import os
from collections import defaultdict

from eyed3.id3 import ID3_V2_4
from eyed3.id3.tag import TagTemplate
from eyed3.plugins import LoaderPlugin
from eyed3.utils import art
from eyed3.utils.prompt import prompt
from eyed3.utils.console import printMsg, Style, Fore
from eyed3 import core

from eyed3.core import (ALBUM_TYPE_IDS, TXXX_ALBUM_TYPE, EP_MAX_SIZE_HINT,
                        LP_TYPE, EP_TYPE, COMP_TYPE, VARIOUS_TYPE, DEMO_TYPE,
                        LIVE_TYPE, SINGLE_TYPE, VARIOUS_ARTISTS)

NORMAL_FNAME_FORMAT = "${artist} - ${track:num} - ${title}"
VARIOUS_FNAME_FORMAT = "${track:num} - ${artist} - ${title}"
SINGLE_FNAME_FORMAT = "${artist} - ${title}"

NORMAL_DNAME_FORMAT = "${best_date:prefer_release} - ${album}"
LIVE_DNAME_FORMAT = "${best_date:prefer_recording} - ${album}"


def _printChecking(msg, end='\n'):
    print(Style.BRIGHT + Fore.GREEN + "Checking" + Style.RESET_ALL + " %s" % msg, end=end)


def _fixCase(s):
    if s:
        fixed_values = []
        for word in s.split():
            fixed_values.append(word.capitalize())
        return " ".join(fixed_values)
    else:
        return s


def dirDate(d):
    s = str(d)
    if "T" in s:
        s = s.split("T")[0]
    return s.replace('-', '.')


class FixupPlugin(LoaderPlugin):
    NAMES = ["fixup"]
    SUMMARY = "Performs various checks and fixes to directories of audio files."
    DESCRIPTION = """
Operates on directories at a time, fixing each as a unit (album,
compilation, live set, etc.). All of these should have common dates,
for example but other characteristics may vary. The ``--type`` should be used
whenever possible, ``lp`` is the default.

The following test and fixes always apply:

    1.  Every file will be given an ID3 tag if one is missing.
    2.  Set ID3 v2.4.
    3.  Set a consistent album name for all files in the directory.
    4.  Set a consistent artist name for all files, unless the type is
        ``various`` in which case the artist may vary (but must exist).
    5.  Ensure each file has a title.
    6.  Ensure each file has a track # and track total.
    7.  Ensure all files have a release and original release date, unless the
        type is ``live`` in which case the recording date is set.
    8.  All ID3 frames of the following types are removed: USER, PRIV
    9.  All ID3 files have TLEN (track length in ms) set (or updated).
    10. The album/dir type is set in the tag. Types of ``lp`` and ``various``
        do not have this field set since the latter is the default and the
        former can be determined during sync. In ID3 terms the value is in
        TXXX (description: ``%(TXXX_ALBUM_TYPE)s``).
    11. Files are renamed as follows:
        - Type ``various``: %(VARIOUS_FNAME_FORMAT)s
        - Type ``single``: %(SINGLE_FNAME_FORMAT)s
        - All other types: %(NORMAL_FNAME_FORMAT)s
        - A rename template can be supplied in --file-rename-pattern
    12. Directories are renamed as follows:
        - Type ``live``: %(LIVE_DNAME_FORMAT)s
        - All other types: %(NORMAL_DNAME_FORMAT)s
        - A rename template can be supplied in --dir-rename-pattern

Album types:

    - ``lp``: A traditinal "album" of songs from a single artist.
      No extra info is written to the tag since this is the default.
    - ``ep``: A short collection of songs from a single artist. The string 'ep'
      is written to the tag's ``%(TXXX_ALBUM_TYPE)s`` field.
    - ``various``: A collection of songs from different artists. The string
      'various' is written to the tag's ``%(TXXX_ALBUM_TYPE)s`` field.
    - ``live``: A collection of live recordings from a single artist. The string
      'live' is written to the tag's ``%(TXXX_ALBUM_TYPE)s`` field.
    - ``compilation``: A collection of songs from various recordings by a single
      artist. The string 'compilation' is written to the tag's
      ``%(TXXX_ALBUM_TYPE)s`` field. Compilation dates, unlike other types, may
      differ.
    - ``demo``: A demo recording by a single artist. The string 'demo' is
      written to the tag's ``%(TXXX_ALBUM_TYPE)s`` field.
    - ``single``: A track that should no be associated with an album (even if
      it has album metadata). The string 'single' is written to the tag's
      ``%(TXXX_ALBUM_TYPE)s`` field.

""" % globals()

    def __init__(self, arg_parser):
        super(FixupPlugin, self).__init__(arg_parser, cache_files=True,
                                          track_images=True)
        g = self.arg_group
        self._handled_one = False

        g.add_argument("--type", choices=ALBUM_TYPE_IDS, dest="dir_type",
                       default=None, help=ARGS_HELP["--type"])
        g.add_argument("--fix-case", action="store_true", dest="fix_case",
                       help=ARGS_HELP["--fix-case"])
        g.add_argument("-n", "--dry-run", action="store_true", dest="dry_run",
                       help=ARGS_HELP["--dry-run"])
        g.add_argument("--no-prompt", action="store_true", dest="no_prompt",
                       help=ARGS_HELP["--no-prompt"])
        g.add_argument("--dotted-dates", action="store_true",
                       help=ARGS_HELP["--dotted-dates"])
        g.add_argument("--file-rename-pattern", dest="file_rename_pattern",
                       help=ARGS_HELP["--file-rename-pattern"])
        g.add_argument("--dir-rename-pattern", dest="dir_rename_pattern",
                       help=ARGS_HELP["--dir-rename-pattern"])
        self._curr_dir_type = None
        self._dir_files_to_remove = set()

    def _getOne(self, key, values, default=None, Type=str,
                required=True):
        values = set(values)
        if None in values:
            values.remove(None)

        if len(values) != 1:
            printMsg(
                "Detected %s %s names%s" %
                ("0" if len(values) == 0 else "multiple",
                 key,
                 "." if not values
                     else (":\n\t%s" % "\n\t".join([str(v) for v in values])),
                ))

            value = prompt("Enter %s" % key.title(), default=default,
                           type_=Type, required=required)
        else:
            value = values.pop()

        return value

    def _getDates(self, audio_files):
        tags = [f.tag for f in audio_files if f.tag]

        rel_dates = set([t.release_date for t in tags if t.release_date])
        orel_dates = set([t.original_release_date for t in tags
                          if t.original_release_date])
        rec_dates = set([t.recording_date for t in tags if t.recording_date])

        release_date, original_release_date, recording_date = None, None, None

        def reduceDate(type_str, dates_set, default_date=None):
            if len(dates_set or []) != 1:
                reduced = self._getOne(type_str, dates_set,
                                       default=str(default_date) if default_date
                                                                 else None,
                                       Type=core.Date.parse)
            else:
                reduced = dates_set.pop()
            return reduced

        if (False not in [a.tag.album_type == LIVE_TYPE for a in audio_files] or
                self._curr_dir_type == LIVE_TYPE):
            # The recording date is most meaningful for live music.
            recording_date = reduceDate("recording date",
                                        rec_dates | orel_dates | rel_dates)
            rec_dates = {recording_date}

            # Want when these set if they may recording time.
            orel_dates.difference_update(rec_dates)
            rel_dates.difference_update(rec_dates)

            if orel_dates:
                original_release_date = reduceDate("original release date",
                                                   orel_dates | rel_dates)
            orel_dates = {original_release_date}

            if rel_dates | orel_dates:
                release_date = reduceDate("release date",
                                          rel_dates | orel_dates)
        elif (False not in [a.tag.album_type == COMP_TYPE
                            for a in audio_files] or
              self._curr_dir_type == COMP_TYPE):
            # The release date is most meaningful for comps, other track dates
            # may differ.
            if len(rel_dates) != 1:
                release_date = reduceDate("release date", rel_dates | orel_dates)
            else:
                release_date = list(rel_dates)[0]
        else:
            if len(orel_dates) != 1:
                # The original release date is most meaningful for studio music.
                original_release_date = reduceDate("original release date",
                                                   orel_dates | rel_dates | rec_dates)
                orel_dates = {original_release_date}
            else:
                original_release_date = list(orel_dates)[0]

            if len(rel_dates) != 1:
                release_date = reduceDate("release date", rel_dates | orel_dates)
                rel_dates = {release_date}
            else:
                release_date = list(rel_dates)[0]

            if rec_dates.difference(orel_dates | rel_dates):
                recording_date = reduceDate("recording date", rec_dates)

        return release_date, original_release_date, recording_date

    def _resolveArtistInfo(self, audio_files):
        assert(self._curr_dir_type != SINGLE_TYPE)

        tags = [f.tag for f in audio_files if f.tag]
        artists = set([t.album_artist for t in tags if t.album_artist])

        # There can be 0 or 1 album artist values.
        album_artist = None
        if len(artists) > 1:
            album_artist = self._getOne("album artist", artists, required=False)
        elif artists:
            album_artist = artists.pop()

        artists = list(set([t.artist for t in tags if t.artist]))

        if len(artists) > 1:
            # There can be more then 1 artist when VARIOUS_TYPE or
            # album_artist != None.
            if not album_artist and self._curr_dir_type != VARIOUS_TYPE:
                if prompt("Multiple artist names exist, process directory as "
                          "various artists", default=True):
                    self._curr_dir_type = VARIOUS_TYPE
                else:
                    artists = [self._getOne("artist", artists, required=True)]
            elif (album_artist == VARIOUS_ARTISTS and
                    self._curr_dir_type != VARIOUS_TYPE):
                self._curr_dir_type = VARIOUS_TYPE
        elif len(artists) == 0:
            artists = [self._getOne("artist", [], required=True)]

        # Fix up artist and album artist discrepancies
        if len(artists) == 1 and album_artist:
            artist = artists[0]
            if album_artist != artist:
                print("When there is only one artist it should match the "
                      "album artist. Choices are: ")
                for s in [artist, album_artist]:
                    print("\t%s" % s)
                album_artist = prompt("Select common artist and album artist",
                                      choices=[artist, album_artist])
                artists = [album_artist]

        if self.args.fix_case:
            album_artist = _fixCase(album_artist)
            artists = [_fixCase(a) for a in artists]
        return album_artist, artists

    def _getAlbum(self, audio_files):
        tags = [f.tag for f in audio_files if f.tag]
        albums = set([t.album for t in tags if t.album])
        album_name = (albums.pop() if len(albums) == 1
                                   else self._getOne("album", albums))
        assert album_name
        return album_name if not self.args.fix_case else _fixCase(album_name)

    def _checkCoverArt(self, directory, audio_files):
        valid_cover = False

        # Check for cover file.
        _printChecking("for cover art...")
        for dimg in self._dir_images:
            art_type = art.matchArtFile(dimg)
            if art_type == art.FRONT_COVER:
                dimg_name = os.path.basename(dimg)
                print("\t%s" % dimg_name)
                valid_cover = True

        if not valid_cover:
            # FIXME: move the logic out fixup and into art.
            #  Look for a cover in the tags.
            for tag in [af.tag for af in audio_files if af.tag]:
                if valid_cover:
                    # It could be set below...
                    break
                for img in tag.images:
                    if img.picture_type == img.FRONT_COVER:
                        file_name = img.makeFileName("cover")
                        print("\tFound front cover in tag, writing '%s'" %
                              file_name)
                        with open(os.path.join(directory, file_name),
                                  "wb") as img_file:
                            img_file.write(img.image_data)
                            img_file.close()
                            valid_cover = True

        return valid_cover

    def start(self, args, config):
        import eyed3.utils.prompt
        eyed3.utils.prompt.DISABLE_PROMPT = "exit" if args.no_prompt else None

        super(FixupPlugin, self).start(args, config)

    def handleFile(self, f, *args, **kwargs):
        super(FixupPlugin, self).handleFile(f, *args, **kwargs)
        if not self.audio_file and f not in self._dir_images:
            self._dir_files_to_remove.add(f)

    def handleDirectory(self, directory, _):
        if not self._file_cache:
            return

        directory = os.path.abspath(directory)
        print("\n" + Style.BRIGHT + Fore.YELLOW +
              "Scanning directory%s %s" % (Style.RESET_ALL, directory))

        def _path(af):
            return af.path

        self._handled_one = True

        # Make sure all of the audio files has a tag.
        for f in self._file_cache:
            if f.tag is None:
                f.initTag()

        audio_files = sorted(list(self._file_cache), key=_path)

        self._file_cache = []
        edited_files = set()
        self._curr_dir_type = self.args.dir_type
        if self._curr_dir_type is None:
            types = set([a.tag.album_type for a in audio_files])
            if len(types) == 1:
                self._curr_dir_type = types.pop()

        # Check for corrections to LP, EP, COMP
        if (self._curr_dir_type is None and
                    len(audio_files) < EP_MAX_SIZE_HINT):
            # Do you want EP?
            if False in [a.tag.album_type == EP_TYPE for a in audio_files]:
                if prompt("Only %d audio files, process directory as an EP" %
                          len(audio_files),
                          default=True):
                    self._curr_dir_type = EP_TYPE
            else:
                self._curr_dir_type = EP_TYPE
        elif (self._curr_dir_type in (EP_TYPE, DEMO_TYPE) and
                len(audio_files) > EP_MAX_SIZE_HINT):
            # Do you want LP?
            if prompt("%d audio files is large for type %s, process "
                      "directory as an LP" % (len(audio_files),
                                              self._curr_dir_type),
                      default=True):
                self._curr_dir_type = LP_TYPE

        last = defaultdict(lambda: None)

        album_artist = None
        artists = set()
        album = None

        if self._curr_dir_type != SINGLE_TYPE:
            album_artist, artists = self._resolveArtistInfo(audio_files)
            print(Fore.BLUE + "Album artist: " + Style.RESET_ALL +
                  (album_artist or ""))
            print(Fore.BLUE + "Artist" + ("s" if len(artists) > 1 else "") +
                  ": " + Style.RESET_ALL + ", ".join(artists))

            album = self._getAlbum(audio_files)
            print(Fore.BLUE + "Album: " + Style.RESET_ALL + album)

            rel_date, orel_date, rec_date = self._getDates(audio_files)
            for what, d in [("Release", rel_date),
                            ("Original", orel_date),
                            ("Recording", rec_date)]:
                print(f"{Fore.BLUE} {what} date: {Style.RESET_ALL} {d}")

            num_audio_files = len(audio_files)
            track_nums = set([f.tag.track_num[0] for f in audio_files])
            fix_track_nums = set(range(1, num_audio_files + 1)) != track_nums
            new_track_nums = []

        dir_type = self._curr_dir_type
        for f in sorted(audio_files, key=_path):
            print(Style.BRIGHT + Fore.GREEN + "Checking" + Fore.RESET +
                  Style.BRIGHT + (" %s" % os.path.basename(f.path)) +
                  Style.RESET_ALL)

            if not f.tag:
                print("\tAdding new tag")
                f.initTag()
                edited_files.add(f)
            tag = f.tag

            if tag.version != ID3_V2_4:
                print("\tConverting to ID3 v2.4")
                tag.version = ID3_V2_4
                edited_files.add(f)

            if dir_type != SINGLE_TYPE and album_artist != tag.album_artist:
                print("\tSetting album artist: %s" % album_artist)
                tag.album_artist = album_artist
                edited_files.add(f)

            if not tag.artist and dir_type in (VARIOUS_TYPE, SINGLE_TYPE):
                # Prompt artist
                tag.artist = prompt("Artist name", default=last["artist"])
                last["artist"] = tag.artist
            elif len(artists) == 1 and tag.artist != artists[0]:
                assert(dir_type != SINGLE_TYPE)
                print("\tSetting artist: %s" % artists[0])
                tag.artist = artists[0]
                edited_files.add(f)

            if tag.album != album and dir_type != SINGLE_TYPE:
                print("\tSetting album: %s" % album)
                tag.album = album
                edited_files.add(f)

            orig_title = tag.title
            if not tag.title:
                tag.title = prompt("Track title")
            tag.title = tag.title.strip()
            if self.args.fix_case:
                tag.title = _fixCase(tag.title)
            if orig_title != tag.title:
                print("\tSetting title: %s" % tag.title)
                edited_files.add(f)

            if dir_type != SINGLE_TYPE:
                # Track numbers
                tnum, ttot = tag.track_num
                update = False
                if ttot != num_audio_files:
                    update = True
                    ttot = num_audio_files

                if fix_track_nums or not (1 <= tnum <= num_audio_files):
                    tnum = None
                    while tnum is None:
                        tnum = int(prompt("Track #", type_=int))
                        if not (1 <= tnum <= num_audio_files):
                            print(Fore.RED + "Out of range: " + Fore.RESET +
                                  "1 <= %d <= %d" % (tnum, num_audio_files))
                            tnum = None
                        elif tnum in new_track_nums:
                            print(Fore.RED + "Duplicate value: " + Fore.RESET +
                                    str(tnum))
                            tnum = None
                        else:
                            update = True
                            new_track_nums.append(tnum)

                if update:
                    tag.track_num = (tnum, ttot)
                    print("\tSetting track numbers: %s" % str(tag.track_num))
                    edited_files.add(f)
            else:
                # Singles
                if tag.track_num != (None, None):
                    tag.track_num = (None, None)
                    edited_files.add(f)

            if dir_type != SINGLE_TYPE:
                # Dates
                if rec_date and tag.recording_date != rec_date:
                    print("\tSetting %s date (%s)" %
                            ("recording", str(rec_date)))
                    tag.recording_date = rec_date
                    edited_files.add(f)
                if rel_date and tag.release_date != rel_date:
                    print("\tSetting %s date (%s)" % ("release", str(rel_date)))
                    tag.release_date = rel_date
                    edited_files.add(f)
                if orel_date and tag.original_release_date != orel_date:
                    print("\tSetting %s date (%s)" % ("original release",
                                                      str(orel_date)))
                    tag.original_release_date = orel_date
                    edited_files.add(f)

            for frame in list(tag.frameiter(["USER", "PRIV"])):
                print("\tRemoving %s frames: %s" %
                        (frame.id,
                         frame.owner_id if frame.id == b"PRIV" else frame.text))
                tag.frame_set[frame.id].remove(frame)
                edited_files.add(f)

            # Add TLEN
            tlen = tag.getTextFrame("TLEN")
            if tlen is not None:
                real_tlen_ms = f.info.time_secs * 1000
                tlen_ms = float(tlen)
                if tlen_ms != real_tlen_ms:
                    print("\tSetting TLEN (%d)" % real_tlen_ms)
                    tag.setTextFrame("TLEN", str(real_tlen_ms))
                    edited_files.add(f)

            # Add custom album type if special and otherwise not able to be
            # determined.
            curr_type = tag.album_type
            if curr_type != dir_type:
                print("\tSetting %s = %s" % (TXXX_ALBUM_TYPE, dir_type))
                tag.album_type = dir_type
                edited_files.add(f)

        try:
            if not self._checkCoverArt(directory, audio_files):
                if not prompt("Proceed without valid cover file", default=True):
                    return
        finally:
            self._dir_images = []

        # Determine other changes, like file and/or directory renames
        # so they can be reported before save confirmation.

        # File renaming
        file_renames = []
        if self.args.file_rename_pattern:
            format_str = self.args.file_rename_pattern
        else:
            if dir_type == SINGLE_TYPE:
                format_str = SINGLE_FNAME_FORMAT
            elif dir_type in (VARIOUS_TYPE, COMP_TYPE):
                format_str = VARIOUS_FNAME_FORMAT
            else:
                format_str = NORMAL_FNAME_FORMAT

        for f in audio_files:
            orig_name, orig_ext = os.path.splitext(os.path.basename(f.path))
            new_name = TagTemplate(format_str).substitute(f.tag, zeropad=True)
            if orig_name != new_name:
                printMsg("Rename file to %s%s" % (new_name, orig_ext))
                file_renames.append((f, new_name, orig_ext))

        # Directory renaming
        dir_rename = None
        if dir_type != SINGLE_TYPE:
            if self.args.dir_rename_pattern:
                dir_format = self.args.dir_rename_pattern
            else:
                if dir_type == LIVE_TYPE:
                    dir_format = LIVE_DNAME_FORMAT
                else:
                    dir_format = NORMAL_DNAME_FORMAT
            template = TagTemplate(dir_format,
                                   dotted_dates=self.args.dotted_dates)

            pref_dir = template.substitute(audio_files[0].tag, zeropad=True)
            if os.path.basename(directory) != pref_dir:
                new_dir = os.path.join(os.path.dirname(directory), pref_dir)
                printMsg("Rename directory to %s" % new_dir)
                dir_rename = (directory, new_dir)

        # Cruft files to remove
        file_removes = []
        if self._dir_files_to_remove:
            for f in self._dir_files_to_remove:
                print("Remove file: " + os.path.basename(f))
                file_removes.append(f)
        self._dir_files_to_remove = set()

        if not self.args.dry_run:
            confirmed = False

            if (edited_files or file_renames or dir_rename or file_removes):
                confirmed = prompt("\nSave changes", default=True)

            if confirmed:
                for f in edited_files:
                    print("Saving %s" % os.path.basename(f.path))
                    f.tag.save(version=ID3_V2_4, preserve_file_time=True)

                for f, new_name, orig_ext in file_renames:
                    printMsg("Renaming file to %s%s" % (new_name, orig_ext))
                    f.rename(new_name, preserve_file_time=True)

                if file_removes:
                    for f in file_removes:
                        printMsg("Removing file %s" % os.path.basename(f))
                        os.remove(f)

                if dir_rename:
                    printMsg("Renaming directory to %s" % dir_rename[1])
                    s = os.stat(dir_rename[0])
                    os.rename(dir_rename[0], dir_rename[1])
                    # With a rename use the origianl access time
                    os.utime(dir_rename[1], (s.st_atime, s.st_atime))

        else:
            printMsg("\nNo changes made (run without -n/--dry-run)")

    def handleDone(self):
        if not self._handled_one:
            printMsg("Nothing to do")


def _getTemplateKeys():
    from eyed3.id3.tag import TagTemplate

    keys = list(TagTemplate("")._makeMapping(None, False).keys())
    keys.sort()
    return ", ".join(["$%s" % v for v in keys])


ARGS_HELP = {
        "--type": "How to treat each directory. The default is '%s', "
                  "although you may be prompted for an alternate choice "
                  "if the files look like another type." % ALBUM_TYPE_IDS[0],
        "--fix-case": "Fix casing on each string field by capitalizing each "
                      "word.",
        "--dry-run": "Only print the operations that would take place, but do "
                      "not execute them.",
        "--no-prompt": "Exit if prompted.",
        "--dotted-dates": "Separate date with '.' instead of '-' when naming "
                          "directories.",
        "--file-rename-pattern": "Rename file (the extension is not affected) "
                                 "based on data in the tag using substitution "
                                 "variables: " + _getTemplateKeys(),
        "--dir-rename-pattern": "Rename directory based on data in the tag "
                                "using substitution variables: " +
                                _getTemplateKeys(),
}



import math
import logging
import binascii
from ..utils import requireBytes
from ..utils.binfuncs import (bin2dec, bytes2bin, bin2bytes,
                              bin2synchsafe, dec2bin)
from .. import core
from . import ID3_DEFAULT_VERSION, isValidVersion, normalizeVersion

from ..utils.log import getLogger
log = getLogger(__name__)

NULL_FRAME_FLAGS = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]


class TagHeader(object):
    SIZE = 10

    def __init__(self, version=ID3_DEFAULT_VERSION):
        self.clear()
        self.version = version

    def clear(self):
        self.tag_size = 0
        # Flag bits
        self.unsync = False
        self.extended = False
        self.experimental = False
        # v2.4 addition
        self.footer = False

    @property
    def version(self):
        return tuple([v for v in self._version])

    @version.setter
    def version(self, v):
        v = normalizeVersion(v)
        if not isValidVersion(v, fully_qualified=True):
            raise ValueError("Invalid version: %s" % str(v))
        self._version = v

    @property
    def major_version(self):
        return self._version[0]

    @property
    def minor_version(self):
        return self._version[1]

    @property
    def rev_version(self):
        return self._version[2]

    def parse(self, f):
        """Parse an ID3 v2 header starting at the current position of `f`.

        If a header is parsed `True` is returned, otherwise `False`. If
        a header is found but malformed an `eyed3.id3.tag.TagException` is
        thrown.
        """
        from .tag import TagException

        self.clear()

        # 3 bytes: v2 header is "ID3".
        if f.read(3) != b"ID3":
            return False
        log.debug("Located ID3 v2 tag")

        # 2 bytes: the minor and revision versions.
        version = f.read(2)
        if len(version) != 2:
            return False
        major = 2
        minor = version[0]
        rev = version[1]
        log.debug("TagHeader [major]: %d " % major)
        log.debug("TagHeader [minor]: %d " % minor)
        log.debug("TagHeader [rev]: %d " % rev)
        if not (major == 2 and (minor >= 2 and minor <= 4)):
            raise TagException("ID3 v%d.%d is not supported" % (major, minor))
        self.version = (major, minor, rev)

        # 1 byte (first 4 bits): flags
        data = f.read(1)
        if not data:
            return False
        (self.unsync,
         self.extended,
         self.experimental,
         self.footer) = (bool(b) for b in bytes2bin(data)[0:4])
        log.debug("TagHeader [flags]: unsync(%d) extended(%d) "
                  "experimental(%d) footer(%d)" % (self.unsync, self.extended,
                                                   self.experimental,
                                                   self.footer))

        # 4 bytes: The size of the extended header (if any), frames, and padding
        # afer unsynchronization. This is a sync safe integer, so only the
        # bottom 7 bits of each byte are used.
        tag_size_bytes = f.read(4)
        if len(tag_size_bytes) != 4:
            return False
        log.debug("TagHeader [size string]: 0x%02x%02x%02x%02x" %
                  (tag_size_bytes[0], tag_size_bytes[1],
                   tag_size_bytes[2], tag_size_bytes[3]))
        self.tag_size = bin2dec(bytes2bin(tag_size_bytes, 7))
        log.debug("TagHeader [size]: %d (0x%x)" % (self.tag_size,
                                                   self.tag_size))

        return True

    def render(self, tag_len=None):
        if tag_len is not None:
            self.tag_size = tag_len

        if self.unsync:
            raise NotImplementedError("eyeD3 does not write (only reads) "
                                      "unsync'd data")

        data = b"ID3"
        data += bytes([self.minor_version]) + bytes([self.rev_version])
        data += bin2bytes([int(self.unsync),
                           int(self.extended),
                           int(self.experimental),
                           int(self.footer),
                           0, 0, 0, 0])
        log.debug("Setting tag size to %d" % self.tag_size)
        data += bin2bytes(bin2synchsafe(dec2bin(self.tag_size, 32)))
        log.debug("TagHeader rendered %d bytes" % len(data))
        return data


class ExtendedTagHeader(object):
    RESTRICT_TAG_SZ_LARGE = 0x00
    RESTRICT_TAG_SZ_MED = 0x01
    RESTRICT_TAG_SZ_SMALL = 0x02
    RESTRICT_TAG_SZ_TINY = 0x03

    RESTRICT_TEXT_ENC_NONE = 0x00
    RESTRICT_TEXT_ENC_UTF8 = 0x01

    RESTRICT_TEXT_LEN_NONE = 0x00
    RESTRICT_TEXT_LEN_1024 = 0x01
    RESTRICT_TEXT_LEN_128 = 0x02
    RESTRICT_TEXT_LEN_30 = 0x03

    RESTRICT_IMG_ENC_NONE = 0x00
    RESTRICT_IMG_ENC_PNG_JPG = 0x01

    RESTRICT_IMG_SZ_NONE = 0x00
    RESTRICT_IMG_SZ_256 = 0x01
    RESTRICT_IMG_SZ_64 = 0x02
    RESTRICT_IMG_SZ_64_EXACT = 0x03

    def __init__(self):
        self.size = 0
        self._flags = 0
        self.crc = None
        self._restrictions = 0

    @property
    def update_bit(self):
        return bool(self._flags & 0x40)

    @update_bit.setter
    def update_bit(self, v):
        if v:
            self._flags |= 0x40
        else:
            self._flags &= ~0x40

    @property
    def crc_bit(self):
        return bool(self._flags & 0x20)

    @crc_bit.setter
    def crc_bit(self, v):
        if v:
            self._flags |= 0x20
        else:
            self._flags &= ~0x20

    @property
    def crc(self):
        return self._crc

    @crc.setter
    def crc(self, v):
        self.crc_bit = 1 if v else 0
        self._crc = v

    @property
    def restrictions_bit(self):
        return bool(self._flags & 0x10)

    @restrictions_bit.setter
    def restrictions_bit(self, v):
        if v:
            self._flags |= 0x10
        else:
            self._flags &= ~0x10

    @property
    def tag_size_restriction(self):
        return self._restrictions >> 6

    @tag_size_restriction.setter
    def tag_size_restriction(self, v):
        assert(v >= 0 and v <= 3)
        self.restrictions_bit = 1
        self._restrictions = (v << 6) | (self._restrictions & 0x3f)

    @property
    def tag_size_restriction_description(self):
        val = self.tag_size_restriction
        if val == ExtendedTagHeader.RESTRICT_TAG_SZ_LARGE:
            return "No more than 128 frames and 1 MB total tag size"
        elif val == ExtendedTagHeader.RESTRICT_TAG_SZ_MED:
            return "No more than 64 frames and 128 KB total tag size"
        elif val == ExtendedTagHeader.RESTRICT_TAG_SZ_SMALL:
            return "No more than 32 frames and 40 KB total tag size"
        elif val == ExtendedTagHeader.RESTRICT_TAG_SZ_TINY:
            return "No more than 32 frames and 4 KB total tag size"

    @property
    def text_enc_restriction(self):
        return (self._restrictions & 0x20) >> 5

    @text_enc_restriction.setter
    def text_enc_restriction(self, v):
        assert(v == 0 or v == 1)
        self.restrictions_bit = 1
        self._restrictions ^= 0x20

    @property
    def text_enc_restriction_description(self):
        if self.text_enc_restriction:
            return "Strings are only encoded with ISO-8859-1 or UTF-8"
        else:
            return "None"

    @property
    def text_length_restriction(self):
        return (self._restrictions >> 3) & 0x03

    @text_length_restriction.setter
    def text_length_restriction(self, v):
        assert(v >= 0 and v <= 3)
        self.restrictions_bit = 1
        self._restrictions = (v << 3) | (self._restrictions & 0xe7)

    @property
    def text_length_restriction_description(self):
        val = self.text_length_restriction
        if val == ExtendedTagHeader.RESTRICT_TEXT_LEN_NONE:
            return "None"
        elif val == ExtendedTagHeader.RESTRICT_TEXT_LEN_1024:
            return "No string is longer than 1024 characters."
        elif val == ExtendedTagHeader.RESTRICT_TEXT_LEN_128:
            return "No string is longer than 128 characters."
        elif val == ExtendedTagHeader.RESTRICT_TEXT_LEN_30:
            return "No string is longer than 30 characters."

    @property
    def image_enc_restriction(self):
        return (self._restrictions & 0x04) >> 2

    @image_enc_restriction.setter
    def image_enc_restriction(self, v):
        assert(v == 0 or v == 1)
        self.restrictions_bit = 1
        self._restrictions ^= 0x04

    @property
    def image_enc_restriction_description(self):
        if self.image_enc_restriction:
            return "Images are encoded only with PNG [PNG] or JPEG [JFIF]."
        else:
            return "None"

    @property
    def image_size_restriction(self):
        return self._restrictions & 0x03

    @image_size_restriction.setter
    def image_size_restriction(self, v):
        assert(v >= 0 and v <= 3)
        self.restrictions_bit = 1
        self._restrictions = v | (self._restrictions & 0xfc)

    @property
    def image_size_restriction_description(self):
        val = self.image_size_restriction
        if val == ExtendedTagHeader.RESTRICT_IMG_SZ_NONE:
            return "None"
        elif val == ExtendedTagHeader.RESTRICT_IMG_SZ_256:
            return "All images are 256x256 pixels or smaller."
        elif val == ExtendedTagHeader.RESTRICT_IMG_SZ_64:
            return "All images are 64x64 pixels or smaller."
        elif val == ExtendedTagHeader.RESTRICT_IMG_SZ_64_EXACT:
            return "All images are exactly 64x64 pixels, unless required "\
                   "otherwise."

    def _syncsafeCRC(self):
        return bytes([
            (self.crc >> 28) & 0x7f,
            (self.crc >> 21) & 0x7f,
            (self.crc >> 14) & 0x7f,
            (self.crc >> 7) & 0x7f,
            (self.crc >> 0) & 0x7f,
        ])

    def render(self, version, frame_data, padding=0):
        assert(version[0] == 2)

        data = b""
        if version[1] == 4:
            # Version 2.4
            size = 6
            # Extended flags.
            if self.update_bit:
                data += b"\x00"
            if self.crc_bit:
                data += b"\x05"
                # XXX: Using the absolute value of the CRC. The spec is unclear
                # about the type of this data.
                self.crc = int(math.fabs(binascii.crc32(frame_data +
                                                        (b"\x00" * padding))))
                crc_data = self._syncsafeCRC()
                if len(crc_data) < 5:
                    # pad if necessary
                    crc_data = (b"\x00" * (5 - len(crc_data))) + crc_data
                assert(len(crc_data) == 5)
                data += crc_data
            if self.restrictions_bit:
                data += b"\x01"
                data += bytes([self._restrictions])
            log.debug("Rendered extended header data (%d bytes)" % len(data))

            # Extended header size.
            size = bin2bytes(bin2synchsafe(dec2bin(len(data) + 6, 32)))
            assert(len(size) == 4)

            data = size + b"\x01" + bin2bytes(dec2bin(self._flags)) + data
            log.debug("Rendered extended header of size %d" % len(data))
        else:
            # Version 2.3
            size = 6  # Note, the 4 size bytes are not included in the size
            # Extended flags.
            f = [0] * 16
            crc = None
            if self.crc_bit:
                f[0] = 1
                # XXX: Using the absolute value of the CRC.  The spec is unclear
                # about the type of this value.
                self.crc = int(math.fabs(binascii.crc32(frame_data +
                                                        (b"\x00" * padding))))
                crc = bin2bytes(dec2bin(self.crc))
                assert(len(crc) == 4)
                size += 4
            flags = bin2bytes(f)
            assert(len(flags) == 2)
            # Extended header size.
            size = bin2bytes(dec2bin(size, 32))
            assert(len(size) == 4)
            # Padding size
            padding_size = bin2bytes(dec2bin(padding, 32))

            data = size + flags + padding_size
            if crc:
                data += crc

        return data

    # Only call this when you *know* there is an extened header.
    def parse(self, fp, version):
        '''Parse an ID3 v2 extended header starting at the current position
        of ``fp`` and per the format defined by ``version``. This method
        should only be called when the presence of an extended header is known
        since it moves the file position. If a header is found but malformed
        an ``eyed3.id3.tag.TagException`` is thrown. The return value is
        ``None``.
        '''
        from .tag import TagException
        assert(version[0] == 2)

        log.debug("Parsing extended header @ 0x%x" % fp.tell())
        # First 4 bytes is the size of the extended header.
        data = fp.read(4)
        if version[1] == 4:
            # sync-safe
            sz = bin2dec(bytes2bin(data, 7))
            self.size = sz
            log.debug("Extended header size (includes the 4 size bytes): %d" %
                      sz)
            data = fp.read(sz - 4)

            # Number of flag bytes
            if data[0] != 1 or (data[1] & 0x8f):
                # As of 2.4 the first byte is 1 and the second can only have
                # bits 6, 5, and 4 set.
                raise TagException("Invalid Extended Header")

            self._flags = data[1]
            log.debug("Extended header flags: %x" % self._flags)

            offset = 2
            if self.update_bit:
                log.debug("Extended header has update bit set")
                assert(data[offset] == 0)
                offset += 1
            if self.crc_bit:
                log.debug("Extended header has CRC bit set")
                assert(data[offset] == 5)
                offset += 1
                crc_data = data[offset:offset + 5]
                # This is sync-safe.
                self.crc = bin2dec(bytes2bin(crc_data, 7))
                log.debug("Extended header CRC: %d" % self.crc)
                offset += 5
            if self.restrictions_bit:
                log.debug("Extended header has restrictions bit set")
                assert(data[offset] == 1)
                offset += 1
                self._restrictions = data[offset]
                offset += 1
        else:
            # v2.3 is totally different... *sigh*
            sz = bin2dec(bytes2bin(data))
            self.size = sz
            log.debug("Extended header size (not including 4 size bytes): %d" %
                      sz)
            tmpFlags = fp.read(2)
            # Read the padding size, but it'll be computed during the parse.
            ps = fp.read(4)
            log.debug("Extended header says there is %d bytes of padding" %
                      bin2dec(bytes2bin(ps)))
            # Make this look like a v2.4 mask.
            self._flags = tmpFlags[0] >> 2
            if self.crc_bit:
                log.debug("Extended header has CRC bit set")
                crc_data = fp.read(4)
                self.crc = bin2dec(bytes2bin(crc_data))
                log.debug("Extended header CRC: %d" % self.crc)


class FrameHeader:
    """A header for each and every ID3 frame in a tag."""

    # 2.4 not only added flag bits, but also reordered the previously defined
    # flags. So these are mapped once the ID3 version is known. Access through
    # 'self', always
    TAG_ALTER = None
    FILE_ALTER = None
    READ_ONLY = None
    COMPRESSED = None
    ENCRYPTED = None
    GROUPED = None
    UNSYNC = None
    DATA_LEN = None

    # Constructor.
    @requireBytes(1)
    def __init__(self, fid, version):
        self._version = version
        self._setBitMask()
        # _setBitMask will throw if the version is no good

        # Correctly set size of header (v2.2 is smaller)
        self.size = 10 if self.minor_version != 2 else 6

        # The frame header itself...
        self.id = fid           # First 4 bytes, frame ID
        self._flags = [0] * 16  # 16 bits, represented here as a list
        self.data_size = 0      # 4 bytes, size of frame data

    def copyFlags(self, rhs):
        self.tag_alter = rhs._flags[rhs.TAG_ALTER]
        self.file_alter = rhs._flags[rhs.FILE_ALTER]
        self.read_only = rhs._flags[rhs.READ_ONLY]
        self.compressed = rhs._flags[rhs.COMPRESSED]
        self.encrypted = rhs._flags[rhs.ENCRYPTED]
        self.grouped = rhs._flags[rhs.GROUPED]
        self.unsync = rhs._flags[rhs.UNSYNC]
        self.data_length_indicator = rhs._flags[rhs.DATA_LEN]

    @property
    def major_version(self):
        return self._version[0]

    @property
    def minor_version(self):
        return self._version[1]

    @property
    def version(self):
        return self._version

    @property
    def tag_alter(self):
        return self._flags[self.TAG_ALTER]

    @tag_alter.setter
    def tag_alter(self, b):
        self._flags[self.TAG_ALTER] = int(bool(b))

    @property
    def file_alter(self):
        return self._flags[self.FILE_ALTER]

    @file_alter.setter
    def file_alter(self, b):
        self._flags[self.FILE_ALTER] = int(bool(b))

    @property
    def read_only(self):
        return self._flags[self.READ_ONLY]

    @read_only.setter
    def read_only(self, b):
        self._flags[self.READ_ONLY] = int(bool(b))

    @property
    def compressed(self):
        return self._flags[self.COMPRESSED]

    @compressed.setter
    def compressed(self, b):
        self._flags[self.COMPRESSED] = int(bool(b))

    @property
    def encrypted(self):
        return self._flags[self.ENCRYPTED]

    @encrypted.setter
    def encrypted(self, b):
        self._flags[self.ENCRYPTED] = int(bool(b))

    @property
    def grouped(self):
        return self._flags[self.GROUPED]

    @grouped.setter
    def grouped(self, b):
        self._flags[self.GROUPED] = int(bool(b))

    @property
    def unsync(self):
        return self._flags[self.UNSYNC]

    @unsync.setter
    def unsync(self, b):
        self._flags[self.UNSYNC] = int(bool(b))

    @property
    def data_length_indicator(self):
        return self._flags[self.DATA_LEN]

    @data_length_indicator.setter
    def data_length_indicator(self, b):
        self._flags[self.DATA_LEN] = int(bool(b))

    def _setBitMask(self):
        major = self.major_version
        minor = self.minor_version

        # 1.x tags are converted to 2.4 frames internally.  These frames are
        # created with frame flags \x00.

        if (major == 2 and minor in (3, 2)):
            # v2.2 does not contain flags, but set anyway, as long as the
            # values remain 0 all is good
            self.TAG_ALTER = 0
            self.FILE_ALTER = 1
            self.READ_ONLY = 2
            self.COMPRESSED = 8
            self.ENCRYPTED = 9
            self.GROUPED = 10
            # This is not in 2.3 frame header flags, map to unused
            self.UNSYNC = 14
            # This is not in 2.3 frame header flags, map to unused
            self.DATA_LEN = 4
        elif ((major == 2 and minor == 4) or (major == 1 and minor in (0, 1))):
            self.TAG_ALTER = 1
            self.FILE_ALTER = 2
            self.READ_ONLY = 3
            self.COMPRESSED = 12
            self.ENCRYPTED = 13
            self.GROUPED = 9
            self.UNSYNC = 14
            self.DATA_LEN = 15
        else:
            raise ValueError("ID3 v" + str(major) + "." + str(minor) +
                             " is not supported.")

    def render(self, data_size):
        data = b''

        assert(type(self.id) is bytes)
        data += self.id

        self.data_size = data_size

        if self.minor_version == 3:
            data += bin2bytes(dec2bin(data_size, 32))
        else:
            data += bin2bytes(bin2synchsafe(dec2bin(data_size, 32)))

        if self.unsync:
            raise NotImplementedError("eyeD3 does not write (only reads) "
                                      "unsync'd data")
        data += bin2bytes(self._flags)

        return data

    @staticmethod
    def _parse2_2(f, version):
        from .frames import map2_2FrameId
        from .frames import FrameException
        frame_id_22 = f.read(3)
        frame_id = map2_2FrameId(frame_id_22)
        if FrameHeader._isValidFrameId(frame_id):
            log.debug("FrameHeader [id]: %s (0x%x%x%x)" %
                      (frame_id_22, frame_id_22[0], frame_id_22[1], frame_id_22[2]))
            frame_header = FrameHeader(frame_id, version)
            # data_size corresponds to the size of the data segment after
            # encryption, compression, and unsynchronization.
            sz = f.read(3)
            frame_header.data_size = bin2dec(bytes2bin(sz, 8))
            log.debug("FrameHeader [data size]: %d (0x%X)" %
                      (frame_header.data_size, frame_header.data_size))
            return frame_header
        elif frame_id == b'\x00\x00\x00':
            log.debug("FrameHeader: Null frame id found at byte %d" % f.tell())
        else:
            core.parseError(FrameException("FrameHeader: Illegal Frame ID: %s" %
                                           frame_id))

        return None

    @staticmethod
    def parse(f, version):
        from .frames import FrameException
        log.debug("FrameHeader [start byte]: %d (0x%X)" % (f.tell(),
                                                           f.tell()))
        major_version, minor_version = version[:2]
        if minor_version == 2:
            return FrameHeader._parse2_2(f, version)

        frame_id = f.read(4)
        if FrameHeader._isValidFrameId(frame_id):
            log.debug("FrameHeader [id]: %s (0x%x%x%x%x)" %
                      (frame_id, frame_id[0], frame_id[1], frame_id[2], frame_id[3]))
            frame_header = FrameHeader(frame_id, version)
            # data_size corresponds to the size of the data segment after
            # encryption, compression, and unsynchronization.
            sz = f.read(4)
            # In ID3 v2.4 this value became a synch-safe integer, meaning only
            # the low 7 bits are used per byte.
            if minor_version == 3:
                frame_header.data_size = bin2dec(bytes2bin(sz, 8))
            else:
                frame_header.data_size = bin2dec(bytes2bin(sz, 7))
            log.debug("FrameHeader [data size]: %d (0x%X)" %
                      (frame_header.data_size, frame_header.data_size))

            # Frame flags.
            flags = f.read(2)
            frame_header._flags = bytes2bin(flags)
            if log.getEffectiveLevel() <= logging.DEBUG:
                log.debug("FrameHeader [flags]: ta(%d) fa(%d) ro(%d) co(%d) "
                          "en(%d) gr(%d) un(%d) dl(%d)" %
                          (frame_header.tag_alter,
                           frame_header.file_alter, frame_header.read_only,
                           frame_header.compressed, frame_header.encrypted,
                           frame_header.grouped, frame_header.unsync,
                           frame_header.data_length_indicator))
            if (frame_header.minor_version >= 4 and frame_header.compressed and
                   not frame_header.data_length_indicator):
                core.parseError(FrameException("Invalid frame; compressed with "
                                               "no data length indicator"))

            return frame_header
        elif frame_id == b'\x00' * 4:
            log.debug("FrameHeader: Null frame id found at byte %d" % f.tell())
        else:
            core.parseError(FrameException("FrameHeader: Illegal Frame ID: %s" %
                                           frame_id))

        return None

    @staticmethod
    def _isValidFrameId(id):
        import re
        return re.compile(b"^[A-Z0-9][A-Z0-9][A-Z0-9][A-Z0-9]$").match(id)



import os
import importlib.machinery
from eyed3.plugins import LoaderPlugin

_DEFAULT_MOD = "eyeD3mod.py"


class PyModulePlugin(LoaderPlugin):
    SUMMARY = "Imports a Python module file and calls its functions for the "\
              "the various plugin events."
    DESCRIPTION = """
If no module if provided a file named %(_DEFAULT_MOD)s in the current working directory is
imported. If any of the following methods exist they still be invoked:

def audioFile(audio_file):
    '''Invoked for every audio file that is encountered. The ``audio_file``
    is of type ``eyed3.core.AudioFile``; currently this is the concrete type
    ``eyed3.mp3.Mp3AudioFile``.'''
    pass

def audioDir(d, audio_files, images):
    '''This function is invoked for any directory (``d``) that contains audio
    (``audio_files``) or image (``images``) media.'''
    pass

def done():
    '''This method is invoke before successful exit.'''
    pass
""" % globals()
    NAMES = ["pymod"]

    def __init__(self, arg_parser):
        super(PyModulePlugin, self).__init__(arg_parser, cache_files=True,
                                             track_images=True)
        self._mod = None
        self.arg_group.add_argument("-m", "--module", dest="module",
                                    help="The Python module module to invoke. "
                                         "The default is ./%s" % _DEFAULT_MOD)

    def start(self, args, config):
        mod_file = args.module or _DEFAULT_MOD
        try:
            mod_name = os.path.splitext(os.path.basename(mod_file))[0]
            loader = importlib.machinery.SourceFileLoader(mod_name, mod_file)
            mod = loader.load_module()
            self._mod = mod
        except IOError:
            raise IOError("Module file not found: %s" % mod_file)
        except (NameError, IndentationError, ImportError, SyntaxError) as ex:
            raise IOError("Module load error: %s" % str(ex))

    def handleFile(self, f):
        super(PyModulePlugin, self).handleFile(f)
        if not self.audio_file:
            return

        if "audioFile" in dir(self._mod):
            self._mod.audioFile(self.audio_file)

    def handleDirectory(self, d, _):
        if not self._file_cache and not self._dir_images:
            return

        if "audioDir" in dir(self._mod):
            self._mod.audioDir(d, self._file_cache, self._dir_images)

        super(PyModulePlugin, self).handleDirectory(d, _)

    def handleDone(self):
        super(PyModulePlugin, self).handleDone()
        if "done" in dir(self._mod):
            self._mod.done()



import deprecation
from math import log10

from . import Mp3Exception
from ..utils.binfuncs import bytes2bin, bytes2dec, bin2dec
from ..utils.log import getLogger
from ..__about__ import __version__

log = getLogger(__name__)


def isValidHeader(header):
    """Determine if ``header`` (an integer, 4 bytes compared) is a valid mp3
    frame header."""
    # Test for the mp3 frame sync: 11 set bits.
    sync = (header >> 16)
    if sync & 0xffe0 != 0xffe0:
        # ffe0 is 11 sync bits, 12 are not used in order to support identifying
        # mpeg v2.5 (bits 20,19)
        return False

    # All the remaining tests are not entirely required, but do help in
    # finding false syncs

    version = (header >> 19) & 0x3
    if version == 1:
        # This is a "reserved" version
        log.debug("invalid mpeg version")
        return False

    layer = (header >> 17) & 0x3
    if layer == 0:
        # This is a "reserved" layer
        log.debug("invalid mpeg layer")
        return False

    bitrate = (header >> 12) & 0xf
    if bitrate in (0, 0xf):
        # free and bad bitrate values
        log.debug("invalid mpeg bitrate")
        return False

    sample_rate = (header >> 10) & 0x3
    if sample_rate == 0x3:
        # this is a "reserved" sample rate
        log.debug("invalid mpeg sample rate")
        return False

    return True


def findHeader(fp, start_pos=0):
    """Locate the first mp3 header in file stream ``fp`` starting a offset
    ``start_pos`` (defaults to 0). Returned is a 3-tuple containing the offset
    where the header was found, the header as an integer, and the header as 4
    bytes. If no header is found header_int will equal 0.
    """

    def isBOM(buffer, pos):
        """Check for unicode BOM"""
        try:
            if pos - 1 >= 0:
                if buffer[pos - 1] == 254:
                    return True
            return buffer[pos + 1] == 254
        except IndexError:
            return False

    def find_sync(_fp, _pos=0):
        chunk_sz = 8192  # Measured as optimal

        _fp.seek(_pos)
        data = _fp.read(chunk_sz)

        while data:
            pos = 0
            while 0 <= pos < chunk_sz:
                pos = data.find(b"\xff", pos)
                if pos == -1:
                    break

                if not isBOM(data, pos):
                    h = data[pos:pos + 4]
                    if len(h) == 4:
                        return tuple([_pos + pos, h])

                pos += 1

            _pos += len(data)
            data = _fp.read(chunk_sz)

        return None, None

    sync_pos, header_bytes = find_sync(fp, start_pos)
    while sync_pos is not None:
        header = bytes2dec(header_bytes)
        if isValidHeader(header):
            return tuple([sync_pos, header, header_bytes])
        sync_pos, header_bytes = find_sync(fp, start_pos + sync_pos + 2)

    return None, None, None


def timePerFrame(mp3_header, vbr):
    """Computes the number of seconds per mp3 frame. It can be used to
    compute overall playtime and bitrate. The mp3 layer and sample
    rate from ``mp3_header`` are used to compute the number of seconds
    (fractional float point value) per mp3 frame. Be sure to set ``vbr`` True
    when dealing with VBR, otherwise playtimes may be incorrect."""
    # https://bitbucket.org/nicfit/eyed3/issue/32/mp3audioinfotime_secs-incorrect-for-mpeg2
    if mp3_header.version >= 2.0 and vbr:
        row = _mp3VersionKey(mp3_header.version)
    else:
        row = 0
    return (float(SAMPLES_PER_FRAME_TABLE[row][mp3_header.layer]) /
            float(mp3_header.sample_freq))


@deprecation.deprecated(deprecated_in="0.9a2", removed_in="1.0", current_version=__version__,
                        details="Use timePerFrame instead")
def compute_time_per_frame(mp3_header):
    if mp3_header is not None:
        return timePerFrame(mp3_header, False)


class Mp3Header:
    """Header container for MP3 frames."""
    def __init__(self, header_data=None):
        self.version = None
        self.layer = None
        self.error_protection = None
        self.bit_rate = None
        self.sample_freq = None
        self.padding = None
        self.private_bit = None
        self.copyright = None
        self.original = None
        self.emphasis = None
        self.mode = None
        # This value is left as is: 0<=mode_extension<=3.
        # See http://www.dv.co.yu/mpgscript/mpeghdr.htm for how to interpret
        self.mode_extension = None
        self.frame_length = None

        if header_data:
            self.decode(header_data)

    # This may throw an Mp3Exception if the header is malformed.
    def decode(self, header):
        if not isValidHeader(header):
            raise Mp3Exception("Invalid MPEG header")

        # MPEG audio version from bits 19 and 20.
        version = (header >> 19) & 0x3
        self.version = [2.5, None, 2.0, 1.0][version]
        if self.version is None:
            raise Mp3Exception("Illegal MPEG version")

        # MPEG layer
        self.layer = 4 - ((header >> 17) & 0x3)
        if self.layer == 4:
            raise Mp3Exception("Illegal MPEG layer")

        # Decode some simple values.
        self.error_protection = not (header >> 16) & 0x1
        self.padding = (header >> 9) & 0x1
        self.private_bit = (header >> 8) & 0x1
        self.copyright = (header >> 3) & 0x1
        self.original = (header >> 2) & 0x1

        # Obtain sampling frequency.
        sample_bits = (header >> 10) & 0x3
        self.sample_freq = \
            SAMPLE_FREQ_TABLE[sample_bits][_mp3VersionKey(self.version)]
        if not self.sample_freq:
            raise Mp3Exception("Illegal MPEG sampling frequency")

        # Compute bitrate.
        bit_rate_row = (header >> 12) & 0xf
        if int(self.version) == 1 and self.layer == 1:
            bit_rate_col = 0
        elif int(self.version) == 1 and self.layer == 2:
            bit_rate_col = 1
        elif int(self.version) == 1 and self.layer == 3:
            bit_rate_col = 2
        elif int(self.version) == 2 and self.layer == 1:
            bit_rate_col = 3
        elif int(self.version) == 2 and (self.layer == 2 or
                                         self.layer == 3):
            bit_rate_col = 4
        else:
            raise Mp3Exception("Mp3 version %f and layer %d is an invalid "
                               "combination" % (self.version, self.layer))
        self.bit_rate = BIT_RATE_TABLE[bit_rate_row][bit_rate_col]
        if self.bit_rate is None:
            raise Mp3Exception("Invalid bit rate")
        # We know know the bit rate specified in this frame, but if the file
        # is VBR we need to obtain the average from the Xing header.
        # This is done by the caller since right now all we have is the frame
        # header.

        # Emphasis; whatever that means??
        emph = header & 0x3
        if emph == 0:
            self.emphasis = EMPHASIS_NONE
        elif emph == 1:
            self.emphasis = EMPHASIS_5015
        elif emph == 2:
            self.emphasis = EMPHASIS_CCIT
        else:
            raise Mp3Exception("Illegal mp3 emphasis value: %d" % emph)

        # Channel mode.
        mode_bits = (header >> 6) & 0x3
        if mode_bits == 0:
            self.mode = MODE_STEREO
        elif mode_bits == 1:
            self.mode = MODE_JOINT_STEREO
        elif mode_bits == 2:
            self.mode = MODE_DUAL_CHANNEL_STEREO
        else:
            self.mode = MODE_MONO
        self.mode_extension = (header >> 4) & 0x3

        # Layer II has restrictions wrt to mode and bit rate.  This code
        # enforces them.
        if self.layer == 2:
            m = self.mode
            br = self.bit_rate
            if (br in [32, 48, 56, 80] and (m != MODE_MONO)):
                raise Mp3Exception("Invalid mode/bitrate combination for layer "
                                   "II")
            if (br in [224, 256, 320, 384] and (m == MODE_MONO)):
                raise Mp3Exception("Invalid mode/bitrate combination for layer "
                                   "II")

        br = self.bit_rate * 1000
        sf = self.sample_freq
        p = self.padding
        if self.layer == 1:
            # Layer 1 uses 32 bit slots for padding.
            p = self.padding * 4
            self.frame_length = int((((12 * br) / sf) + p) * 4)
        else:
            # Layer 2 and 3 uses 8 bit slots for padding.
            p = self.padding * 1
            self.frame_length = int(((144 * br) / sf) + p)

        # Dump the state.
        log.debug("MPEG audio version: " + str(self.version))
        log.debug("MPEG audio layer: " + ("I" * self.layer))
        log.debug("MPEG sampling frequency: " + str(self.sample_freq))
        log.debug("MPEG bit rate: " + str(self.bit_rate))
        log.debug("MPEG channel mode: " + self.mode)
        log.debug("MPEG channel mode extension: " + str(self.mode_extension))
        log.debug("MPEG CRC error protection: " + str(self.error_protection))
        log.debug("MPEG original: " + str(self.original))
        log.debug("MPEG copyright: " + str(self.copyright))
        log.debug("MPEG private bit: " + str(self.private_bit))
        log.debug("MPEG padding: " + str(self.padding))
        log.debug("MPEG emphasis: " + str(self.emphasis))
        log.debug("MPEG frame length: " + str(self.frame_length))


class VbriHeader(object):
    def __init__(self):
        self.vbr = True
        self.version = None

    ##
    # \brief Decode the VBRI info from \a frame.
    # http://www.codeproject.com/audio/MPEGAudioInfo.asp#VBRIHeader
    def decode(self, frame):

        # The header is 32 bytes after the end of the first MPEG audio header,
        # therefore 4 + 32 = 36
        offset = 36
        head = frame[offset:offset + 4]
        if head != 'VBRI':
            return False
        log.debug("VBRI header detected @ %x" % (offset))
        offset += 4

        self.version = bin2dec(bytes2bin(frame[offset:offset + 2]))
        offset += 2

        self.delay = bin2dec(bytes2bin(frame[offset:offset + 2]))
        offset += 2

        self.quality = bin2dec(bytes2bin(frame[offset:offset + 2]))
        offset += 2

        self.num_bytes = bin2dec(bytes2bin(frame[offset:offset + 4]))
        offset += 4

        self.num_frames = bin2dec(bytes2bin(frame[offset:offset + 4]))
        offset += 4

        return True


class XingHeader:
    """Header class for the Xing header extensions."""

    def __init__(self):
        self.numFrames = int()
        self.numBytes = int()
        self.toc = [0] * 100
        self.vbrScale = int()

    # Pass in the first mp3 frame from the file as a byte string.
    # If an Xing header is present in the file it'll be in the first mp3
    # frame.  This method returns true if the Xing header is found in the
    # frame, and false otherwise.
    def decode(self, frame):
        # mp3 version
        version = (frame[1] >> 3) & 0x1
        # channel mode.
        mode = (frame[3] >> 6) & 0x3

        # Find the start of the Xing header.
        if version:
            # +4 in all of these to skip initial mp3 frame header.
            if mode != 3:
                pos = 32 + 4
            else:
                pos = 17 + 4
        else:
            if mode != 3:
                pos = 17 + 4
            else:
                pos = 9 + 4
        head = frame[pos:pos + 4]
        self.vbr = (head == b'Xing') and True or False
        if head not in [b'Xing', b'Info']:
            return False
        log.debug("%s header detected @ %x" % (head, pos))
        pos += 4

        # Read Xing flags.
        headFlags = bin2dec(bytes2bin(frame[pos:pos + 4]))
        pos += 4
        log.debug("%s header flags: 0x%x" % (head, headFlags))

        # Read frames header flag and value if present
        if headFlags & FRAMES_FLAG:
            self.numFrames = bin2dec(bytes2bin(frame[pos:pos + 4]))
            pos += 4
            log.debug("%s numFrames: %d" % (head, self.numFrames))

        # Read bytes header flag and value if present
        if headFlags & BYTES_FLAG:
            self.numBytes = bin2dec(bytes2bin(frame[pos:pos + 4]))
            pos += 4
            log.debug("%s numBytes: %d" % (head, self.numBytes))

        # Read TOC header flag and value if present
        if headFlags & TOC_FLAG:
            self.toc = frame[pos:pos + 100]
            pos += 100
            log.debug("%s TOC (100 bytes): PRESENT" % head)
        else:
            log.debug("%s TOC (100 bytes): NOT PRESENT" % head)

        # Read vbr scale header flag and value if present
        if headFlags & VBR_SCALE_FLAG and head == b'Xing':
            self.vbrScale = bin2dec(bytes2bin(frame[pos:pos + 4]))
            pos += 4
            log.debug("%s vbrScale: %d" % (head, self.vbrScale))

        return True


class LameHeader(dict):
    r""" Mp3 Info tag (AKA LAME Tag)

    Lame (and some other encoders) write a tag containing various bits of info
    about the options used at encode time.  If available, the following are
    parsed and stored in the LameHeader dict:

    encoder_version: short encoder version [str]
    tag_revision:    revision number of the tag [int]
    vbr_method:      VBR method used for encoding [str]
    lowpass_filter:  lowpass filter frequency in Hz [int]
    replaygain:      if available, radio and audiofile gain (see below) [dict]
    encoding_flags:  encoding flags used [list]
    nogap:           location of gaps when --nogap was used [list]
    ath_type:        ATH type [int]
    bitrate:         bitrate and type (Constant, Target, Minimum) [tuple]
    encoder_delay:   samples added at the start of the mp3 [int]
    encoder_padding: samples added at the end of the mp3 [int]
    noise_shaping:   noise shaping method [int]
    stereo_mode:     stereo mode used [str]
    unwise_settings: whether unwise settings were used [boolean]
    sample_freq:     source sample frequency [str]
    mp3_gain:        mp3 gain adjustment (rarely used) [float]
    preset:          preset used [str]
    surround_info:   surround information [str]
    music_length:    length in bytes of original mp3 [int]
    music_crc:       CRC-16 of the mp3 music data [int]
    infotag_crc:     CRC-16 of the info tag [int]

    Prior to ~3.90, Lame simply stored the encoder version in the first frame.
    If the infotag_crc is invalid, then we try to read this version string.  A
    simple way to tell if the LAME Tag is complete is to check for the
    infotag_crc key.

    Replay Gain data is only available since Lame version 3.94b.  If set, the
    replaygain dict has the following structure:

    \code
       peak_amplitude: peak signal amplitude [float]
       radio:
          name:       name of the gain adjustment [str]
          adjustment: gain adjustment [float]
          originator: originator of the gain adjustment [str]
       audiofile: [same as radio]
    \endcode

    Note that as of 3.95.1, Lame uses 89dB as a reference level instead of the
    83dB that is specified in the Replay Gain spec.  This is not automatically
    compensated for.  You can do something like this if you want:

    \code
       import eyeD3
       af = eyeD3.mp3.Mp3AudioFile('/path/to/some.mp3')
       lamever = af.lameTag['encoder_version']
       name, ver = lamever[:4], lamever[4:]
       gain = af.lameTag['replaygain']['radio']['adjustment']
       if name == 'LAME' and eyeD3.mp3.lamevercmp(ver, '3.95') > 0:
           gain -= 6
    \endcode

    Radio and Audiofile Replay Gain are often referrered to as Track and Album
    gain, respectively.  See http://replaygain.hydrogenaudio.org/ for futher
    details on Replay Gain.

    See http://gabriel.mp3-tech.org/mp3infotag.html for the gory details of the
    LAME Tag.
    """

    # from the LAME source:
    # http://lame.cvs.sourceforge.net/*checkout*/lame/lame/libmp3lame/VbrTag.c
    _crc16_table = [
      0x0000, 0xC0C1, 0xC181, 0x0140, 0xC301, 0x03C0, 0x0280, 0xC241,
      0xC601, 0x06C0, 0x0780, 0xC741, 0x0500, 0xC5C1, 0xC481, 0x0440,
      0xCC01, 0x0CC0, 0x0D80, 0xCD41, 0x0F00, 0xCFC1, 0xCE81, 0x0E40,
      0x0A00, 0xCAC1, 0xCB81, 0x0B40, 0xC901, 0x09C0, 0x0880, 0xC841,
      0xD801, 0x18C0, 0x1980, 0xD941, 0x1B00, 0xDBC1, 0xDA81, 0x1A40,
      0x1E00, 0xDEC1, 0xDF81, 0x1F40, 0xDD01, 0x1DC0, 0x1C80, 0xDC41,
      0x1400, 0xD4C1, 0xD581, 0x1540, 0xD701, 0x17C0, 0x1680, 0xD641,
      0xD201, 0x12C0, 0x1380, 0xD341, 0x1100, 0xD1C1, 0xD081, 0x1040,
      0xF001, 0x30C0, 0x3180, 0xF141, 0x3300, 0xF3C1, 0xF281, 0x3240,
      0x3600, 0xF6C1, 0xF781, 0x3740, 0xF501, 0x35C0, 0x3480, 0xF441,
      0x3C00, 0xFCC1, 0xFD81, 0x3D40, 0xFF01, 0x3FC0, 0x3E80, 0xFE41,
      0xFA01, 0x3AC0, 0x3B80, 0xFB41, 0x3900, 0xF9C1, 0xF881, 0x3840,
      0x2800, 0xE8C1, 0xE981, 0x2940, 0xEB01, 0x2BC0, 0x2A80, 0xEA41,
      0xEE01, 0x2EC0, 0x2F80, 0xEF41, 0x2D00, 0xEDC1, 0xEC81, 0x2C40,
      0xE401, 0x24C0, 0x2580, 0xE541, 0x2700, 0xE7C1, 0xE681, 0x2640,
      0x2200, 0xE2C1, 0xE381, 0x2340, 0xE101, 0x21C0, 0x2080, 0xE041,
      0xA001, 0x60C0, 0x6180, 0xA141, 0x6300, 0xA3C1, 0xA281, 0x6240,
      0x6600, 0xA6C1, 0xA781, 0x6740, 0xA501, 0x65C0, 0x6480, 0xA441,
      0x6C00, 0xACC1, 0xAD81, 0x6D40, 0xAF01, 0x6FC0, 0x6E80, 0xAE41,
      0xAA01, 0x6AC0, 0x6B80, 0xAB41, 0x6900, 0xA9C1, 0xA881, 0x6840,
      0x7800, 0xB8C1, 0xB981, 0x7940, 0xBB01, 0x7BC0, 0x7A80, 0xBA41,
      0xBE01, 0x7EC0, 0x7F80, 0xBF41, 0x7D00, 0xBDC1, 0xBC81, 0x7C40,
      0xB401, 0x74C0, 0x7580, 0xB541, 0x7700, 0xB7C1, 0xB681, 0x7640,
      0x7200, 0xB2C1, 0xB381, 0x7340, 0xB101, 0x71C0, 0x7080, 0xB041,
      0x5000, 0x90C1, 0x9181, 0x5140, 0x9301, 0x53C0, 0x5280, 0x9241,
      0x9601, 0x56C0, 0x5780, 0x9741, 0x5500, 0x95C1, 0x9481, 0x5440,
      0x9C01, 0x5CC0, 0x5D80, 0x9D41, 0x5F00, 0x9FC1, 0x9E81, 0x5E40,
      0x5A00, 0x9AC1, 0x9B81, 0x5B40, 0x9901, 0x59C0, 0x5880, 0x9841,
      0x8801, 0x48C0, 0x4980, 0x8941, 0x4B00, 0x8BC1, 0x8A81, 0x4A40,
      0x4E00, 0x8EC1, 0x8F81, 0x4F40, 0x8D01, 0x4DC0, 0x4C80, 0x8C41,
      0x4400, 0x84C1, 0x8581, 0x4540, 0x8701, 0x47C0, 0x4680, 0x8641,
      0x8201, 0x42C0, 0x4380, 0x8341, 0x4100, 0x81C1, 0x8081, 0x4040]

    ENCODER_FLAGS = {
      'NSPSYTUNE': 0x0001,
      'NSSAFEJOINT': 0x0002,
      'NOGAP_NEXT': 0x0004,
      'NOGAP_PREV': 0x0008}

    PRESETS = {
      0: 'Unknown',
      # 8 to 320 are reserved for ABR bitrates
      410: 'V9',
      420: 'V8',
      430: 'V7',
      440: 'V6',
      450: 'V5',
      460: 'V4',
      470: 'V3',
      480: 'V2',
      490: 'V1',
      500: 'V0',
      1000: 'r3mix',
      1001: 'standard',
      1002: 'extreme',
      1003: 'insane',
      1004: 'standard/fast',
      1005: 'extreme/fast',
      1006: 'medium',
      1007: 'medium/fast'}

    REPLAYGAIN_NAME = {
      0: 'Not set',
      1: 'Radio',
      2: 'Audiofile'}

    REPLAYGAIN_ORIGINATOR = {
      0: 'Not set',
      1: 'Set by artist',
      2: 'Set by user',
      3: 'Set automatically',
      100: 'Set by simple RMS average'}

    SAMPLE_FREQUENCIES = {
      0: '<= 32 kHz',
      1: '44.1 kHz',
      2: '48 kHz',
      3: '> 48 kHz'}

    STEREO_MODES = {
      0: 'Mono',
      1: 'Stereo',
      2: 'Dual',
      3: 'Joint',
      4: 'Force',
      5: 'Auto',
      6: 'Intensity',
      7: 'Undefined'}

    SURROUND_INFO = {
      0: 'None',
      1: 'DPL encoding',
      2: 'DPL2 encoding',
      3: 'Ambisonic encoding',
      8: 'Reserved'}

    VBR_METHODS = {
      0: 'Unknown',
      1: 'Constant Bitrate',
      2: 'Average Bitrate',
      3: 'Variable Bitrate method1 (old/rh)',
      4: 'Variable Bitrate method2 (mtrh)',
      5: 'Variable Bitrate method3 (mt)',
      6: 'Variable Bitrate method4',
      8: 'Constant Bitrate (2 pass)',
      9: 'Average Bitrate (2 pass)',
      15: 'Reserved'}

    def __init__(self, frame):
        """Read the LAME info tag.
        frame should be the first frame of an mp3.
        """
        super().__init__()
        self.decode(frame)

    def _crc16(self, data, val=0):
        """Compute a CRC-16 checksum on a data stream."""
        for c in [bytes([b]) for b in data]:
            val = self._crc16_table[ord(c) ^ (val & 0xff)] ^ (val >> 8)
        return val

    def decode(self, frame):
        """Decode the LAME info tag."""
        try:
            pos = frame.index(b"LAME")
        except ValueError:
            return

        log.debug(f"Lame info tag found at position {pos}")

        # check the info tag crc. If it's not valid, no point parsing much more.
        lamecrc = bin2dec(bytes2bin(frame[190:192]))
        if self._crc16(frame[:190]) != lamecrc:
            log.warning("Lame tag CRC check failed")
        else:
            log.debug("Lame tag CRC OK")

        try:
            # Encoder short VersionString, 9 bytes
            self['encoder_version'] = str(frame[pos:pos + 9].rstrip(), "latin1")
            log.debug('Lame Encoder Version: %s' % self['encoder_version'])
            pos += 9

            # Info Tag revision + VBR method, 1 byte
            self['tag_revision'] = bin2dec(bytes2bin(frame[pos:pos + 1])[:5])
            vbr_method = bin2dec(bytes2bin(frame[pos:pos + 1])[5:])
            self['vbr_method'] = self.VBR_METHODS.get(vbr_method, 'Unknown')
            log.debug('Lame info tag version: %s' % self['tag_revision'])
            log.debug('Lame VBR method: %s' % self['vbr_method'])
            pos += 1

            # Lowpass filter value, 1 byte
            self['lowpass_filter'] = bin2dec(
                                       bytes2bin(frame[pos:pos + 1])) * 100
            log.debug('Lame Lowpass filter value: %s Hz' %
                      self['lowpass_filter'])
            pos += 1

            # Replay Gain, 8 bytes total
            replaygain = {}

            # Peak signal amplitude, 4 bytes
            peak = bin2dec(bytes2bin(frame[pos:pos + 4])) << 5
            if peak > 0:
                peak /= float(1 << 28)
                db = 20 * log10(peak)
                replaygain['peak_amplitude'] = peak
                log.debug('Lame Peak signal amplitude: %.8f (%+.1f dB)' %
                          (peak, db))
            pos += 4

            # Radio and Audiofile Gain, AKA track and album, 2 bytes each
            for gaintype in ['radio', 'audiofile']:
                name = bin2dec(bytes2bin(frame[pos:pos + 2])[:3])
                orig = bin2dec(bytes2bin(frame[pos:pos + 2])[3:6])
                sign = bin2dec(bytes2bin(frame[pos:pos + 2])[6:7])
                adj = bin2dec(bytes2bin(frame[pos:pos + 2])[7:]) / 10.0
                if sign:
                    adj *= -1

                # Lame 3.95.1 and above use 89dB as a reference instead of 83dB as defined by the
                # Replay Gain spec. This will be compensated for with `adj -= 6`
                lamever = self['encoder_version']
                if lamever[:4] == 'LAME' and lamevercmp(lamever[4:], "3.95") > 0:
                    adj -= 6

                if orig:
                    name = self.REPLAYGAIN_NAME.get(name, 'Unknown')
                    orig = self.REPLAYGAIN_ORIGINATOR.get(orig, 'Unknown')
                    replaygain[gaintype] = {'name': name, 'adjustment': adj,
                                            'originator': orig}
                    log.debug('Lame %s Replay Gain: %s dB (%s)' %
                              (name, adj, orig))
                pos += 2
            if replaygain:
                self['replaygain'] = replaygain

            # Encoding flags + ATH Type, 1 byte
            encflags = bin2dec(bytes2bin(frame[pos:pos + 1])[:4])
            (self['encoding_flags'],
             self['nogap']) = self._parse_encflags(encflags)
            self['ath_type'] = bin2dec(bytes2bin(frame[pos:pos + 1])[4:])
            log.debug('Lame Encoding flags: %s' %
                      ' '.join(self['encoding_flags']))
            if self['nogap']:
                log.debug('Lame No gap: %s' % ' and '.join(self['nogap']))
            log.debug('Lame ATH type: %s' % self['ath_type'])
            pos += 1

            # if ABR {specified bitrate} else {minimal bitrate}, 1 byte
            btype = 'Constant'
            if 'Average' in self['vbr_method']:
                btype = 'Target'
            elif 'Variable' in self['vbr_method']:
                btype = 'Minimum'
            # bitrate may be modified below after preset is read
            self['bitrate'] = (bin2dec(bytes2bin(frame[pos:pos + 1])), btype)
            log.debug('Lame Bitrate (%s): %s' % (btype, self['bitrate'][0]))
            pos += 1

            # Encoder delays, 3 bytes
            self['encoder_delay'] = bin2dec(bytes2bin(frame[pos:pos + 3])[:12])
            self['encoder_padding'] = bin2dec(
                                        bytes2bin(frame[pos:pos + 3])[12:])
            log.debug('Lame Encoder delay: %s samples' % self['encoder_delay'])
            log.debug('Lame Encoder padding: %s samples' %
                      self['encoder_padding'])
            pos += 3

            # Misc, 1 byte
            sample_freq = bin2dec(bytes2bin(frame[pos:pos + 1])[:2])
            unwise_settings = bin2dec(bytes2bin(frame[pos:pos + 1])[2:3])
            stereo_mode = bin2dec(bytes2bin(frame[pos:pos + 1])[3:6])
            self['noise_shaping'] = bin2dec(bytes2bin(frame[pos:pos + 1])[6:])
            self['sample_freq'] = self.SAMPLE_FREQUENCIES.get(sample_freq,
                                                              'Unknown')
            self['unwise_settings'] = bool(unwise_settings)
            self['stereo_mode'] = self.STEREO_MODES.get(stereo_mode, 'Unknown')
            log.debug('Lame Source Sample Frequency: %s' % self['sample_freq'])
            log.debug('Lame Unwise settings used: %s' % self['unwise_settings'])
            log.debug('Lame Stereo mode: %s' % self['stereo_mode'])
            log.debug('Lame Noise Shaping: %s' % self['noise_shaping'])
            pos += 1

            # MP3 Gain, 1 byte
            sign = bytes2bin(frame[pos:pos + 1])[0]
            gain = bin2dec(bytes2bin(frame[pos:pos + 1])[1:])
            if sign:
                gain *= -1
            self['mp3_gain'] = gain
            db = gain * 1.5
            log.debug('Lame MP3 Gain: %s (%+.1f dB)' % (self['mp3_gain'], db))
            pos += 1

            # Preset and surround info, 2 bytes
            surround = bin2dec(bytes2bin(frame[pos:pos + 2])[2:5])
            preset = bin2dec(bytes2bin(frame[pos:pos + 2])[5:])
            if preset in range(8, 321):
                if self['bitrate'][0] >= 255:
                    # the value from preset is better in this case
                    self['bitrate'] = (preset, btype)
                    log.debug('Lame Bitrate (%s): %s' %
                              (btype, self['bitrate'][0]))
                if 'Average' in self['vbr_method']:
                    preset = 'ABR %s' % preset
                else:
                    preset = 'CBR %s' % preset
            else:
                preset = self.PRESETS.get(preset, preset)
            self['surround_info'] = self.SURROUND_INFO.get(surround, surround)
            self['preset'] = preset
            log.debug('Lame Surround Info: %s' % self['surround_info'])
            log.debug('Lame Preset: %s' % self['preset'])
            pos += 2

            # MusicLength, 4 bytes
            self['music_length'] = bin2dec(bytes2bin(frame[pos:pos + 4]))
            log.debug('Lame Music Length: %s bytes' % self['music_length'])
            pos += 4

            # MusicCRC, 2 bytes
            self['music_crc'] = bin2dec(bytes2bin(frame[pos:pos + 2]))
            log.debug('Lame Music CRC: %04X' % self['music_crc'])
            pos += 2

            # CRC-16 of Info Tag, 2 bytes
            self['infotag_crc'] = lamecrc  # we read this earlier
            log.debug('Lame Info Tag CRC: %04X' % self['infotag_crc'])
            pos += 2
        except IndexError:
            log.warning("Truncated LAME info header, values incomplete.")

    def _parse_encflags(self, flags):
        """Parse encoder flags.

        Returns a tuple containing lists of encoder flags and nogap data in
        human readable format.
        """

        encoder_flags, nogap = [], []

        if not flags:
            return encoder_flags, nogap

        if flags & self.ENCODER_FLAGS['NSPSYTUNE']:
            encoder_flags.append('--nspsytune')
        if flags & self.ENCODER_FLAGS['NSSAFEJOINT']:
            encoder_flags.append('--nssafejoint')

        NEXT = self.ENCODER_FLAGS['NOGAP_NEXT']
        PREV = self.ENCODER_FLAGS['NOGAP_PREV']
        if flags & (NEXT | PREV):
            encoder_flags.append('--nogap')
            if flags & PREV:
                nogap.append('before')
            if flags & NEXT:
                nogap.append('after')
        return encoder_flags, nogap


def lamevercmp(x, y):
    """Compare LAME version strings.

    alpha and beta versions are considered older.
    Versions with sub minor parts or end with 'r' are considered newer.

    :param x: The first version to compare.
    :param y: The second version to compare.
    :returns: Return negative if x<y, zero if x==y, positive if x>y.
    """

    def cmp(a, b):
        # This is Python2's built-in `cmp`, which was removed from Python3
        # And depends on bool - bool yielding the integer -1, 0, 1
        return (a > b) - (a < b)

    x = x.ljust(5)
    y = y.ljust(5)
    if x[:5] == y[:5]:
        return 0
    ret = cmp(x[:4], y[:4])
    if ret:
        return ret
    xmaj, xmin = x.split('.')[:2]
    ymaj, ymin = y.split('.')[:2]
    minparts = ['.']
    # lame 3.96.1 added the use of r in the very short version for post releases
    if (xmaj == '3' and xmin >= '96') or (ymaj == '3' and ymin >= '96'):
        minparts.append('r')
    if x[4] in minparts:
        return 1
    if y[4] in minparts:
        return -1
    if x[4] == ' ':
        return 1
    if y[4] == ' ':
        return -1
    return cmp(x[4], y[4])


#                     MPEG1  MPEG2  MPEG2.5
SAMPLE_FREQ_TABLE = ((44100, 22050, 11025),
                     (48000, 24000, 12000),
                     (32000, 16000, 8000),
                     (None, None, None))

#              V1/L1  V1/L2 V1/L3 V2/L1 V2/L2&L3
BIT_RATE_TABLE = ((0,    0,    0,    0,    0),                          # noqa
                  (32,   32,   32,   32,   8),                          # noqa
                  (64,   48,   40,   48,   16),                         # noqa
                  (96,   56,   48,   56,   24),                         # noqa
                  (128,  64,   56,   64,   32),                         # noqa
                  (160,  80,   64,   80,   40),                         # noqa
                  (192,  96,   80,   96,   48),                         # noqa
                  (224,  112,  96,   112,  56),                         # noqa
                  (256,  128,  112,  128,  64),                         # noqa
                  (288,  160,  128,  144,  80),                         # noqa
                  (320,  192,  160,  160,  96),                         # noqa
                  (352,  224,  192,  176,  112),                        # noqa
                  (384,  256,  224,  192,  128),                        # noqa
                  (416,  320,  256,  224,  144),                        # noqa
                  (448,  384,  320,  256,  160),                        # noqa
                  (None, None, None, None, None))

# Rows 1 and 2 (mpeg 2.x) are only used for those versions *and* VBR.
#                                  L1   L2   L3
SAMPLES_PER_FRAME_TABLE = ((None, 384, 1152, 1152),  # MPEG 1
                           (None, 384, 1152, 576),   # MPEG 2
                           (None, 384, 1152, 576),   # MPEG 2.5
                          )

# Emphasis constants
EMPHASIS_NONE = "None"
EMPHASIS_5015 = "50/15 ms"
EMPHASIS_CCIT = "CCIT J.17"

# Mode constants
MODE_STEREO = "Stereo"
MODE_JOINT_STEREO = "Joint stereo"
MODE_DUAL_CHANNEL_STEREO = "Dual channel stereo"
MODE_MONO = "Mono"

# Xing flag bits
FRAMES_FLAG = 0x0001
BYTES_FLAG = 0x0002
TOC_FLAG = 0x0004
VBR_SCALE_FLAG = 0x0008


def _mp3VersionKey(version):
    """Map mp3 version float to a data structure index.
    1 -> 0, 2 -> 1, 2.5 -> 2
    """
    key = None
    if version == 2.5:
        key = 2
    else:
        key = int(version - 1)
    assert(0 <= key <= 2)
    return key



"""
Here lies Apple frames, all of which are non-standard. All of these would have
been standard user text frames by anyone not being a bastard, on purpose.
"""
from .frames import Frame, TextFrame

PCST_FID = b"PCST"
WFED_FID = b"WFED"
TKWD_FID = b"TKWD"
TDES_FID = b"TDES"
TGID_FID = b"TGID"
GRP1_FID = b"GRP1"


class PCST(Frame):
    """Indicates a podcast. The 4 bytes of data is undefined, and is typically all 0."""

    def __init__(self, _=None):
        super().__init__(PCST_FID)

    def render(self):
        self.data = b"\x00" * 4
        return super(PCST, self).render()


class TKWD(TextFrame):
    """Podcast keywords."""

    def __init__(self, _=None, **kwargs):
        super().__init__(TKWD_FID, **kwargs)


class TDES(TextFrame):
    """Podcast description. One encoding byte followed by text per encoding."""

    def __init__(self, _=None, **kwargs):
        super().__init__(TDES_FID, **kwargs)


class TGID(TextFrame):
    """Podcast URL of the audio file. This should be a W frame!"""

    def __init__(self, _=None, **kwargs):
        super().__init__(TGID_FID, **kwargs)


class WFED(TextFrame):
    """Another podcast URL, the feed URL it is said."""

    def __init__(self, _=None, url=""):
        super().__init__(WFED_FID, url)


class GRP1(TextFrame):
    """Apple grouping, could be a TIT1 conversion."""

    def __init__(self, _=None, **kwargs):
        super().__init__(GRP1_FID, **kwargs)



import os
import re
import math
import pathlib
import logging
import argparse
import warnings
import functools

import deprecation

from ..utils.log import getLogger
from .. import LOCAL_FS_ENCODING
from ..__about__ import __version__, __release_name__, __version_txt__

if hasattr(os, "fwalk"):
    os_walk = functools.partial(os.fwalk, follow_symlinks=True)

    def os_walk_unpack(w):
        return w[0:3]

else:
    os_walk = functools.partial(os.walk, followlinks=True)

    def os_walk_unpack(w):
        return w

log = getLogger(__name__)


@deprecation.deprecated(deprecated_in="0.9a2", removed_in="1.0", current_version=__version__,
                        details="Use eyed3.mimetype.guessMimetype() instead.")
def guessMimetype(filename, with_encoding=False):
    from .. import mimetype

    retval = mimetype.guessMimetype(filename)

    if not with_encoding:
        return retval
    else:
        warnings.warn("File character encoding no longer returned, value is None",
                      UserWarning, stacklevel=2)
        return retval, None


def walk(handler, path, excludes=None, fs_encoding=LOCAL_FS_ENCODING, recursive=False):
    """A wrapper around os.walk which handles exclusion patterns and multiple
    path types (str, pathlib.Path, bytes).
    """
    if isinstance(path, pathlib.Path):
        path = str(path)
    else:
        path = str(path, fs_encoding) if type(path) is not str else path

    excludes = excludes if excludes else []
    excludes_re = []
    for e in excludes:
        excludes_re.append(re.compile(e))

    def _isExcluded(_p):
        for ex in excludes_re:
            match = ex.match(_p)
            if match:
                return True
        return False

    if not os.path.exists(path):
        raise IOError(f"file not found: {path}")
    elif os.path.isfile(path) and not _isExcluded(path):
        # If not given a directory, invoke the handler and return
        handler.handleFile(os.path.abspath(path))
        return

    for root, dirs, files in [os_walk_unpack(w) for w in os_walk(path)]:
        root = root if type(root) is str else str(root, fs_encoding)
        dirs.sort()
        files.sort()
        for f in list(files):
            f_key = f
            f = f if type(f) is str else str(f, fs_encoding)
            f = os.path.abspath(os.path.join(root, f))

            if not os.path.isfile(f) or _isExcluded(f):
                files.remove(f_key)
                continue

            try:
                handler.handleFile(f)
            except StopIteration:
                return

        if files:
            handler.handleDirectory(root, files)

        if not recursive:
            break


class FileHandler(object):
    """A handler interface for :func:`eyed3.utils.walk` callbacks."""

    def handleFile(self, f):
        """Called for each file walked. The file ``f`` is the full path and
        the return value is ignored. If the walk should abort the method should
        raise a ``StopIteration`` exception."""
        pass

    def handleDirectory(self, d, files):
        """Called for each directory ``d`` **after** ``handleFile`` has been
        called for each file in ``files``. ``StopIteration`` may be raised to
        halt iteration."""
        pass

    def handleDone(self):
        """Called when there are no more files to handle."""
        pass


def _requireArgType(arg_type, *args):
    arg_indices = []
    kwarg_names = []
    for a in args:
        if type(a) is int:
            arg_indices.append(a)
        else:
            kwarg_names.append(a)
    assert(arg_indices or kwarg_names)

    def wrapper(fn):
        def wrapped_fn(*args, **kwargs):
            for i in arg_indices:
                if i >= len(args):
                    # The ith argument is not there, as in optional arguments
                    break
                if args[i] is not None and not isinstance(args[i], arg_type):
                    raise TypeError("%s(argument %d) must be %s" %
                                    (fn.__name__, i, str(arg_type)))
            for name in kwarg_names:
                if (name in kwargs and kwargs[name] is not None and
                        not isinstance(kwargs[name], arg_type)):
                    raise TypeError("%s(argument %s) must be %s" %
                                    (fn.__name__, name, str(arg_type)))
            return fn(*args, **kwargs)
        return wrapped_fn
    return wrapper


def requireUnicode(*args):
    """Function decorator to enforce str/unicode argument types.
    ``None`` is a valid argument value, in all cases, regardless of not being
    unicode.  ``*args`` Positional arguments may be numeric argument index
    values (requireUnicode(1, 3) - requires argument 1 and 3 are unicode)
    or keyword argument names (requireUnicode("title")) or a combination
    thereof.
    """
    return _requireArgType(str, *args)


def requireBytes(*args):
    """Function decorator to enforce byte string argument types.
    """
    return _requireArgType(bytes, *args)


def formatTime(seconds, total=None, short=False):
    """
    Format ``seconds`` (number of seconds) as a string representation.
    When ``short`` is False (the default) the format is:

        HH:MM:SS.

    Otherwise, the format is exacly 6 characters long and of the form:

        1w 3d
        2d 4h
        1h 5m
        1m 4s
        15s

    If ``total`` is not None it will also be formatted and
    appended to the result seperated by ' / '.
    """
    seconds = round(seconds)

    def time_tuple(ts):
        if ts is None or ts < 0:
            ts = 0
        hours = ts / 3600
        mins = (ts % 3600) / 60
        secs = (ts % 3600) % 60
        tstr = '%02d:%02d' % (mins, secs)
        if int(hours):
            tstr = '%02d:%s' % (hours, tstr)
        return (int(hours), int(mins), int(secs), tstr)

    if not short:
        hours, mins, secs, curr_str = time_tuple(seconds)
        retval = curr_str
        if total:
            hours, mins, secs, total_str = time_tuple(total)
            retval += ' / %s' % total_str
        return retval
    else:
        units = [
            ('y', 60 * 60 * 24 * 7 * 52),
            ('w', 60 * 60 * 24 * 7),
            ('d', 60 * 60 * 24),
            ('h', 60 * 60),
            ('m', 60),
            ('s', 1),
        ]

        seconds = int(seconds)

        if seconds < 60:
            return '   {0:02d}s'.format(seconds)
        for i in range(len(units) - 1):
            unit1, limit1 = units[i]
            unit2, limit2 = units[i + 1]
            if seconds >= limit1:
                return '{0:02d}{1}{2:02d}{3}'.format(
                    seconds // limit1, unit1,
                    (seconds % limit1) // limit2, unit2)
        return '  ~inf'


# Number of bytes per KB (2^10)
KB_BYTES = 1024
# Number of bytes per MB (2^20)
MB_BYTES = 1048576
# Number of bytes per GB (2^30)
GB_BYTES = 1073741824
# Kilobytes abbreviation
KB_UNIT = "KB"
# Megabytes abbreviation
MB_UNIT = "MB"
# Gigabytes abbreviation
GB_UNIT = "GB"


def formatSize(size, short=False):
    """Format ``size`` (nuber of bytes) into string format doing KB, MB, or GB
    conversion where necessary.

    When ``short`` is False (the default) the format is smallest unit of
    bytes and largest gigabytes; '234 GB'.
    The short version is 2-4 characters long and of the form

        256b
        64k
        1.1G
    """
    if not short:
        unit = "Bytes"
        if size >= GB_BYTES:
            size = float(size) / float(GB_BYTES)
            unit = GB_UNIT
        elif size >= MB_BYTES:
            size = float(size) / float(MB_BYTES)
            unit = MB_UNIT
        elif size >= KB_BYTES:
            size = float(size) / float(KB_BYTES)
            unit = KB_UNIT
        return "%.2f %s" % (size, unit)
    else:
        suffixes = ' kMGTPEH'
        if size == 0:
            num_scale = 0
        else:
            num_scale = int(math.floor(math.log(size) / math.log(1000)))
        if num_scale > 7:
            suffix = '?'
        else:
            suffix = suffixes[num_scale]
        num_scale = int(math.pow(1000, num_scale))
        value = size / num_scale
        str_value = str(value)
        if len(str_value) >= 3 and str_value[2] == '.':
            str_value = str_value[:2]
        else:
            str_value = str_value[:3]
        return "{0:>3s}{1}".format(str_value, suffix)


def formatTimeDelta(td):
    """Format a timedelta object ``td`` into a string. """
    days = td.days
    hours = td.seconds / 3600
    mins = (td.seconds % 3600) / 60
    secs = (td.seconds % 3600) % 60
    tstr = "%02d:%02d:%02d" % (hours, mins, secs)
    if days:
        tstr = "%d days %s" % (days, tstr)
    return tstr


def chunkCopy(src_fp, dest_fp, chunk_sz=(1024 * 512)):
    """Copy ``src_fp`` to ``dest_fp`` in ``chunk_sz`` byte increments."""
    done = False
    while not done:
        data = src_fp.read(chunk_sz)
        if data:
            dest_fp.write(data)
        else:
            done = True
        del data


class ArgumentParser(argparse.ArgumentParser):
    """Subclass of argparse.ArgumentParser that adds version and log level
    options."""

    def __init__(self, *args, **kwargs):
        from eyed3 import version as VERSION
        from eyed3.utils.log import LEVELS
        from eyed3.utils.log import MAIN_LOGGER

        def pop_kwarg(name, default):
            if name in kwargs:
                value = kwargs.pop(name) or default
            else:
                value = default
            return value
        main_logger = pop_kwarg("main_logger", MAIN_LOGGER)
        version = pop_kwarg("version", VERSION)

        self.log_levels = [logging.getLevelName(l).lower() for l in LEVELS]

        formatter = argparse.RawDescriptionHelpFormatter
        super(ArgumentParser, self).__init__(*args, formatter_class=formatter,
                                             **kwargs)

        self.add_argument("--version", action="version", version=version,
                          help="Display version information and exit")
        self.add_argument("--about", action="store_true", dest="about_eyed3",
                          help="Display full version, release name, additional info, and exit")

        debug_group = self.add_argument_group("Debugging")
        debug_group.add_argument(
                "-l", "--log-level", metavar="LEVEL[:LOGGER]",
                action=LoggingAction, main_logger=main_logger,
                help="Set a log level. This option may be specified multiple "
                     "times. If a logger name is specified than the level "
                     "applies only to that logger, otherwise the level is set "
                     "on the top-level logger. Acceptable levels are %s. " %
                     (", ".join("'%s'" % l for l in self.log_levels)))
        debug_group.add_argument("--profile", action="store_true",
                                 default=False, dest="debug_profile",
                       help="Run using python profiler.")
        debug_group.add_argument("--pdb", action="store_true", dest="debug_pdb",
                                 help="Drop into 'pdb' when errors occur.")

    def parse_args(self, *args, **kwargs):
        args = super().parse_args(*args, **kwargs)
        if "about_eyed3" in args and args.about_eyed3:
            action = [a for a in self._actions if isinstance(a, argparse._VersionAction)][0]
            version = action.version
            release_name = f" {__release_name__}" if __release_name__ else ""
            print(f"{version}{release_name}\n\n{__version_txt__}")
            self.exit()
        else:
            return args


class LoggingAction(argparse._AppendAction):
    def __init__(self, *args, **kwargs):
        self.main_logger = kwargs.pop("main_logger")
        super(LoggingAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        values = values.split(':')
        level, logger = values if len(values) > 1 else (values[0],
                                                        self.main_logger)

        logger = logging.getLogger(logger)
        try:
            logger.setLevel(logging._nameToLevel[level.upper()])
        except KeyError:
            msg = f"invalid level choice: {level} (choose from {parser.log_levels})"
            raise argparse.ArgumentError(self, msg)

        super(LoggingAction, self).__call__(parser, namespace, values, option_string)


def datePicker(thing, prefer_recording_date=False):
    """This function returns a date of some sort, amongst all the possible
    dates (members called release_date, original_release_date,
    and recording_date of type eyed3.core.Date).

    The order of preference is:
    1) date of original release
    2) date of this versions release
    3) the recording date.

    Unless ``prefer_recording_date`` is ``True`` in which case the order is
    3, 1, 2.

    ``None`` will be returned if no dates are available."""
    if not prefer_recording_date:
        return (thing.original_release_date or
                thing.release_date or
                thing.recording_date)
    else:
        return (thing.recording_date or
                thing.original_release_date or
                thing.release_date)


def makeUniqueFileName(file_path, uniq=''):
    """The ``file_path`` is the desired file name, and it is returned if the
    file does not exist. In the case that it already exists the path is
    adjusted to be unique. First, the ``uniq`` string is added, and then
    a couter is used to find a unique name."""

    path = os.path.dirname(file_path)
    file = os.path.basename(file_path)
    name, ext = os.path.splitext(file)
    count = 1
    while os.path.exists(os.path.join(path, file)):
        if uniq:
            name = "%s_%s" % (name, uniq)
            file = "".join([name, ext])
            uniq = ''
        else:
            file = "".join(["%s_%s" % (name, count), ext])
            count += 1
    return os.path.join(path, file)


def b(x, encoder=None):
    """Converts `x` to a bytes string if not already.
    :param x: The string.
    :param encoder: Optional codec encoder to perform the conversion. The default is
                    `codecs.latin_1_encode`.
    :return: The byte string if conversion was needed.
    """
    if isinstance(x, bytes):
        return x
    else:
        import codecs
        encoder = encoder or codecs.latin_1_encode
        return encoder(x)[0]



"""
~~~~~~~~~~ DO NOT EDIT THIS FILE! Autogenerated by `regarding` ~~~~~~~~~~
https://github.com/nicfit/regarding
"""
import dataclasses

__all__ = ["Version", "project_name", "version", "version_info", "release_name",
           "author", "author_email", "years", "description", "homepage"]


@dataclasses.dataclass
class Version:
    major: int
    minor: int
    maint: int
    release: str
    release_name: str


project_name = "eyeD3"
version = "0.9.6"
release_name = "True Blue"
author = "Travis Shirk"
author_email = "travis@pobox.com"
years = "2002-2020"
version_info = Version(
    0, 9, 6,
    "final", "True Blue"
)
description = "Python audio data toolkit (ID3 and MP3)"
homepage = ""



import re
import functools

from .. import core
from .. import Error
from ..utils.log import getLogger

log = getLogger(__name__)

# Version 1, 1.0 or 1.1
ID3_V1 = (1, None, None)
# Version 1.0, specifically
ID3_V1_0 = (1, 0, 0)
# Version 1.1, specifically
ID3_V1_1 = (1, 1, 0)
# Version 2, 2.2, 2.3 or 2.4
ID3_V2 = (2, None, None)
# Version 2.2, specifically
ID3_V2_2 = (2, 2, 0)
# Version 2.3, specifically
ID3_V2_3 = (2, 3, 0)
# Version 2.4, specifically
ID3_V2_4 = (2, 4, 0)
# The default version for eyeD3 tags and save operations.
ID3_DEFAULT_VERSION = ID3_V2_4
# Useful for operations where any version will suffice.
ID3_ANY_VERSION = (ID3_V1[0] | ID3_V2[0], None, None)

# Byte code for latin1
LATIN1_ENCODING = b"\x00"
# Byte code for UTF-16
UTF_16_ENCODING = b"\x01"
# Byte code for UTF-16 (big endian)
UTF_16BE_ENCODING = b"\x02"
# Byte code for UTF-8 (Not supported in ID3 versions < 2.4)
UTF_8_ENCODING = b"\x03"

# Default language code for frames that contain a language portion.
DEFAULT_LANG = b"eng"

ID3_MIME_TYPE = "application/x-id3"
ID3_MIME_TYPE_EXTENSIONS = (".id3", ".tag")


def isValidVersion(v, fully_qualified=False):
    """Check the tuple ``v`` against the list of valid ID3 version constants.
    If ``fully_qualified`` is ``True`` it is enforced that there are 3
    components to the version in ``v``. Returns ``True`` when valid and
    ``False`` otherwise."""
    valid = v in [ID3_V1, ID3_V1_0, ID3_V1_1,
                  ID3_V2, ID3_V2_2, ID3_V2_3, ID3_V2_4,
                  ID3_ANY_VERSION]
    if not valid:
        return False

    if fully_qualified:
        return None not in (v[0], v[1], v[2])
    else:
        return True


def normalizeVersion(v):
    """If version tuple ``v`` is of the non-specific type (v1 or v2, any, etc.)
    a fully qualified version is returned."""
    if v == ID3_V1:
        v = ID3_V1_1
    elif v == ID3_V2:
        assert ID3_DEFAULT_VERSION[0] & ID3_V2[0]
        v = ID3_DEFAULT_VERSION
    elif v == ID3_ANY_VERSION:
        v = ID3_DEFAULT_VERSION

    # Now, correct bogus version as seen in the wild
    if v[:2] == (2, 2) and v[2] != 0:
        v = (2, 2, 0)

    return v


# Convert an ID3 version constant to a display string
def versionToString(v):
    """Conversion version tuple ``v`` to a string description."""
    if v == ID3_ANY_VERSION:
        return "v1.x/v2.x"
    elif v[0] == 1:
        if v == ID3_V1_0:
            return "v1.0"
        elif v == ID3_V1_1:
            return "v1.1"
        elif v == ID3_V1:
            return "v1.x"
    elif v[0] == 2:
        if v == ID3_V2_2:
            return "v2.2"
        elif v == ID3_V2_3:
            return "v2.3"
        elif v == ID3_V2_4:
            return "v2.4"
        elif v == ID3_V2:
            return "v2.x"
    raise ValueError("Invalid ID3 version constant: %s" % str(v))


class GenreException(Error):
    """Excpetion type for exceptions related to genres."""


@functools.total_ordering
class Genre:
    """A genre in terms of a ``name`` and and ``id``. Only when ``name`` is
    a "standard" genre (as defined by ID3 v1) will ``id`` be a value other
    than ``None``."""

    def __init__(self, name=None, id: int=None, genre_map=None):
        """Constructor takes an optional name and ID. If `id` is
        provided the `name`, regardless of value, is set to the string the
        id maps to. Likewise, if `name` is passed and is a standard genre the
        id is set to the correct value. Any invalid id values cause a
        `ValueError` to be raised. Genre names that are not in the standard
        list are still accepted but the `id` value is set to `None`."""
        self._id, self._name = None, None
        self._genre_map = genre_map or genres
        if not name and id is None:
            return

        # An ID always takes precedence
        if id is not None:
            try:
                self.id = id
                # valid id will set name
                if name and name != self.name:
                    log.warning(f"Genre ID takes precedence and remapped '{name}' to '{self.name}'")
            except ValueError:
                log.warning(f"Invalid numeric genre ID: {id}")
                if not name:
                    # Gave an invalid ID and no name to fallback on
                    raise
                self.name = name
                self.id = None
        else:
            # All we have is a name
            self.name = name

        assert self.id or self.name

    @property
    def id(self):
        """The Genre's id property.
        When setting the value is strictly enforced and if the value is not
        a valid genre code a ``ValueError`` is raised. Otherwise the id is
        set **and** the ``name`` property is updated to the code's string
        name.
        """
        return self._id

    @id.setter
    def id(self, val):
        if val is None:
            self._id = None
            return

        val = int(val)
        if val not in self._genre_map.keys() or not self._genre_map[val]:
            raise ValueError(f"Unknown genre ID: {val}")

        name = self._genre_map[val]
        self._id = val
        self._name = name

    @property
    def name(self):
        """The Genre's name property.
        When setting the value the name is looked up in the standard genre
        map and if found the ``id`` ppropery is set to the numeric valud **and**
        the name is normalized to the sting found in the map. Non standard
        genres are set (with a warning log) and the ``id`` is set to ``None``.
        It is valid to set the value to ``None``.
        """
        return self._name

    @name.setter
    def name(self, val):
        if val is None:
            self._name = None
            return

        if val.lower() in list(self._genre_map.keys()):
            self._id = self._genre_map[val]
            # normalize the name
            self._name = self._genre_map[self._id]
        else:
            log.warning(f"Non standard genre name: {val}")
            self._id = None
            self._name = val

    @staticmethod
    def parse(g_str, id3_std=True):
        """Parses genre information from `genre_str`.
        The following formats are supported:
        01, 2, 23, 125 - ID3 v1.x style.
        (01), (2), (129)Hardcore, (9)Metal, Indie - ID3v2 style with and without
                                                    refinement.
        Raises GenreException when an invalid string is passed.
        """

        g_str = g_str.strip()
        if not g_str:
            return None

        def strip0Padding(s):
            if len(s) > 1:
                return s.lstrip("0")
            else:
                return s

        if id3_std:
            # ID3 v1 style.
            # Match 03, 34, 129.
            if re.compile(r"[0-9][0-9]*$").match(g_str):
                return Genre(id=int(strip0Padding(g_str)))

            # ID3 v2 style.
            # Match (03), (0)Blues, (15) Rap
            v23_match = re.compile(r"\(([0-9][0-9]*)\)(.*)$").match(g_str)
            if v23_match:
                (gid, name) = v23_match.groups()

                gid = int(strip0Padding(gid))
                if gid and name:
                    gid = gid
                    name = name.strip()
                else:
                    gid = gid
                    name = None

                return Genre(id=gid, name=name)

        return Genre(id=None, name=g_str)

    def __str__(self):
        s = ""
        if self.id is not None:
            s += f"({self.id:d})"
        if self.name:
            s += self.name
        return s

    def __eq__(self, rhs):
        if not rhs:
            return False
        elif type(rhs) is str:
            return self.name == rhs
        else:
            return self.id == rhs.id and self.name == rhs.name

    def __lt__(self, rhs):
        if not rhs:
            return False
        elif type(rhs) is str:
            return self.name == rhs
        else:
            return self.name < rhs.name


class GenreMap(dict):
    """Classic genres defined around ID3 v1 but suitable anywhere.  This class
    is used primarily as a way to map numeric genre values to a string name.
    Genre strings on the other hand are not required to exist in this list.
    """
    GENRE_MIN = 0
    GENRE_MAX = None
    ID3_GENRE_MIN = 0
    ID3_GENRE_MAX = 79
    WINAMP_GENRE_MIN = 80
    WINAMP_GENRE_MAX = 191
    GENRE_ID3V1_MAX = 255

    def __init__(self, *args):
        """The optional ``*args`` are passed directly to the ``dict``
        constructor."""
        global ID3_GENRES
        super().__init__(*args)

        # ID3 genres as defined by the v1.1 spec with WinAmp extensions.
        for i, g in enumerate(ID3_GENRES):
            self[i] = g
            self[g.lower() if g else None] = i

        GenreMap.GENRE_MAX = len(ID3_GENRES) - 1
        # Pad up to 255
        for i in range(GenreMap.GENRE_MAX + 1, 255 + 1):
            self[i] = None
        self[None] = 255

    def get(self, key):
        if type(key) is int:
            name, gid = self[key], key
        else:
            gid = self[key]
            name = self[gid]
        return Genre(name, id=gid, genre_map=self)

    def __getitem__(self, key):
        if key and type(key) is not int:
            key = key.lower()
        return super().__getitem__(key)

    @property
    def ids(self):
        return list(sorted([k for k in self.keys() if type(k) is int and self[k]]))

    def iter(self):
        for gid in self.ids:
            g = self[gid]
            if g:
                yield Genre(g, id=gid)


class TagFile(core.AudioFile):
    """
    A shim class for dealing with files that contain only ID3 data, no audio.
    """
    def __init__(self, path, version=ID3_ANY_VERSION):
        self._tag_version = version
        core.AudioFile.__init__(self, path)
        assert(self.type == core.AUDIO_NONE)

    def _read(self):

        with open(self.path, 'rb') as file_obj:
            tag = Tag()
            tag_found = tag.parse(file_obj, self._tag_version)
            self._tag = tag if tag_found else None

        self.type = core.AUDIO_NONE

    def initTag(self, version=ID3_DEFAULT_VERSION):
        """Add a id3.Tag to the file (removing any existing tag if one exists).
        """
        self.tag = Tag()
        self.tag.version = version
        self.tag.file_info = FileInfo(self.path)


# ID3 genres, as defined in ID3 v1. The position in the list is the genre's numeric byte value.
ID3_GENRES = [
    'Blues',
    'Classic Rock',
    'Country',
    'Dance',
    'Disco',
    'Funk',
    'Grunge',
    'Hip-Hop',
    'Jazz',
    'Metal',
    'New Age',
    'Oldies',
    'Other',
    'Pop',
    'R&B',
    'Rap',
    'Reggae',
    'Rock',
    'Techno',
    'Industrial',
    'Alternative',
    'Ska',
    'Death Metal',
    'Pranks',
    'Soundtrack',
    'Euro-Techno',
    'Ambient',
    'Trip-Hop',
    'Vocal',
    'Jazz+Funk',
    'Fusion',
    'Trance',
    'Classical',
    'Instrumental',
    'Acid',
    'House',
    'Game',
    'Sound Clip',
    'Gospel',
    'Noise',
    'AlternRock',
    'Bass',
    'Soul',
    'Punk',
    'Space',
    'Meditative',
    'Instrumental Pop',
    'Instrumental Rock',
    'Ethnic',
    'Gothic',
    'Darkwave',
    'Techno-Industrial',
    'Electronic',
    'Pop-Folk',
    'Eurodance',
    'Dream',
    'Southern Rock',
    'Comedy',
    'Cult',
    'Gangsta Rap',
    'Top 40',
    'Christian Rap',
    'Pop / Funk',
    'Jungle',
    'Native American',
    'Cabaret',
    'New Wave',
    'Psychedelic',
    'Rave',
    'Showtunes',
    'Trailer',
    'Lo-Fi',
    'Tribal',
    'Acid Punk',
    'Acid Jazz',
    'Polka',
    'Retro',
    'Musical',
    'Rock & Roll',
    'Hard Rock',
    'Folk',
    'Folk-Rock',
    'National Folk',
    'Swing',
    'Fast Fusion',
    'Bebob',
    'Latin',
    'Revival',
    'Celtic',
    'Bluegrass',
    'Avantgarde',
    'Gothic Rock',
    'Progressive Rock',
    'Psychedelic Rock',
    'Symphonic Rock',
    'Slow Rock',
    'Big Band',
    'Chorus',
    'Easy Listening',
    'Acoustic',
    'Humour',
    'Speech',
    'Chanson',
    'Opera',
    'Chamber Music',
    'Sonata',
    'Symphony',
    'Booty Bass',
    'Primus',
    'Porn Groove',
    'Satire',
    'Slow Jam',
    'Club',
    'Tango',
    'Samba',
    'Folklore',
    'Ballad',
    'Power Ballad',
    'Rhythmic Soul',
    'Freestyle',
    'Duet',
    'Punk Rock',
    'Drum Solo',
    'A Cappella',
    'Euro-House',
    'Dance Hall',
    'Goa',
    'Drum & Bass',
    'Club-House',
    'Hardcore',
    'Terror',
    'Indie',
    'BritPop',
    'Negerpunk',
    'Polsk Punk',
    'Beat',
    'Christian Gangsta Rap',
    'Heavy Metal',
    'Black Metal',
    'Crossover',
    'Contemporary Christian',
    'Christian Rock',
    'Merengue',
    'Salsa',
    'Thrash Metal',
    'Anime',
    'JPop',
    'Synthpop',
    # https://de.wikipedia.org/wiki/Liste_der_ID3v1-Genres
    'Abstract',
    'Art Rock',
    'Baroque',
    'Bhangra',
    'Big Beat',
    'Breakbeat',
    'Chillout',
    'Downtempo',
    'Dub',
    'EBM',
    'Eclectic',
    'Electro',
    'Electroclash',
    'Emo',
    'Experimental',
    'Garage',
    'Global',
    'IDM',
    'Illbient',
    'Industro-Goth',
    'Jam Band',
    'Krautrock',
    'Leftfield',
    'Lounge',
    'Math Rock',
    'New Romantic',
    'Nu-Breakz',
    'Post-Punk',
    'Post-Rock',
    'Psytrance',
    'Shoegaze',
    'Space Rock',
    'Trop Rock',
    'World Music',
    'Neoclassical',
    'Audiobook',
    'Audio Theatre',
    'Neue Deutsche Welle',
    'Podcast',
    'Indie Rock',
    'G-Funk',
    'Dubstep',
    'Garage Rock',
    'Psybient',
]

# A map of standard genre names and IDs per the ID3 v1 genre definition.
genres = GenreMap()

from . import frames                                                   # noqa
from .tag import Tag, TagException, TagTemplate, FileInfo              # noqa



import math
from eyed3 import id3
from eyed3.plugins import Plugin


class GenreListPlugin(Plugin):
    SUMMARY = "Display the full list of standard ID3 genres."
    DESCRIPTION = "ID3 v1 defined a list of genres and mapped them to "\
                  "to numeric values so they can be stored as a single "\
                  "byte.\nIt is *recommended* that these genres are used "\
                  "although most newer software (including eyeD3) does not "\
                  "care."
    NAMES = ["genres"]

    def __init__(self, arg_parser):
        super(GenreListPlugin, self).__init__(arg_parser)
        self.arg_group.add_argument("-1", "--single-column", action="store_true",
                                    help="List on genre per line.")

    def start(self, args, config):
        self._printGenres(args)

    @staticmethod
    def _printGenres(args):
        # Filter out 'Unknown'
        genre_ids = [i for i in id3.genres
                        if type(i) is int and id3.genres[i] is not None]
        genre_ids.sort()

        if args.single_column:
            for gid in genre_ids:
                print("%3d: %s" % (gid, id3.genres[gid]))
        else:
            offset = int(math.ceil(float(len(genre_ids)) / 2))
            for i in range(offset):
                if i < len(genre_ids):
                    c1 = "%3d: %s" % (i, id3.genres[i])
                else:
                    c1 = ""
                if (i * 2) < len(genre_ids):
                    try:
                        c2 = "%3d: %s" % (i + offset, id3.genres[i + offset])
                    except IndexError:
                        break
                else:
                    c2 = ""
                print(c1 + (" " * (40 - len(c1))) + c2)
            print("")



import os
import sys
import textwrap
import warnings
import deprecation

from io import StringIO
from configparser import ConfigParser
from configparser import Error as ConfigParserError

import eyed3
import eyed3.utils
import eyed3.utils.console
import eyed3.plugins
import eyed3.__about__

from eyed3.utils.log import initLogging

DEFAULT_PLUGIN = "classic"
DEFAULT_CONFIG = os.path.expandvars("${HOME}/.config/eyeD3/config.ini")
USER_PLUGINS_DIR = os.path.expandvars("${HOME}/.config/eyeD3/plugins")
DEFAULT_CONFIG_DEPRECATED = os.path.expandvars("${HOME}/.eyeD3/config.ini")
USER_PLUGINS_DIR_DEPRECATED = os.path.expandvars("${HOME}/.eyeD3/plugins")


def main(args, config):
    if "list_plugins" in args and args.list_plugins:
        _listPlugins(config)
        return 0

    args.plugin.start(args, config)

    recursive = False
    if "non_recursive" in args:
        recursive = not args.non_recursive
    elif "recursive" in args:
        recursive = args.recursive

    # Process paths (files/directories)
    for p in args.paths:
        eyed3.utils.walk(args.plugin, p, excludes=args.excludes, fs_encoding=args.fs_encoding,
                         recursive=recursive)

    retval = args.plugin.handleDone()

    return retval or 0


def _listPlugins(config):
    from eyed3.utils.console import Fore, Style

    def header(name):
        is_default = name == DEFAULT_PLUGIN
        return (Style.BRIGHT + (Fore.GREEN if is_default else '') + "* " +
                name + Style.RESET_ALL)

    all_plugins = eyed3.plugins.load(reload=True, paths=_getPluginPath(config))
    # Create a new dict for sorted display
    plugin_names = []
    for plugin in set(all_plugins.values()):
        plugin_names.append(plugin.NAMES[0])

    print("\nType 'eyeD3 --plugin=<name> --help' for more help\n")

    plugin_names.sort()
    for name in plugin_names:
        plugin = all_plugins[name]

        alt_names = plugin.NAMES[1:]
        alt_names = f" ({', '.join(alt_names)})" if alt_names else ""

        print(f"{header(name)} {alt_names}:")
        for txt in textwrap.wrap(plugin.SUMMARY, initial_indent=' ' * 2, subsequent_indent=' ' * 2):
            print(f"{Fore.YELLOW}{txt}{Style.RESET_ALL}")
        print("")


@deprecation.deprecated(deprecated_in="0.9a2", removed_in="1.0",
                        current_version=eyed3.__about__.__version__,
                        details=f"Default eyeD3 config moved to {DEFAULT_CONFIG}")
def _deprecatedConfigFileCheck(_):
    """This here to add deprecation."""


def _loadConfig(args):
    config_files = []

    if args.config:
        config_files.append(os.path.abspath(args.config))

    if args.no_config is False:
        config_files.append(DEFAULT_CONFIG)
        config_files.append(DEFAULT_CONFIG_DEPRECATED)

    if not config_files:
        return None

    for config_file in config_files:
        if os.path.isfile(config_file):
            _deprecatedConfigFileCheck(config_file)

            try:
                config = ConfigParser()
                config.read(config_file)
            except ConfigParserError as ex:
                eyed3.log.warning(f"User config error: {ex}")
                return None
            else:
                return config
        elif config_file != DEFAULT_CONFIG and config_file != DEFAULT_CONFIG_DEPRECATED:
            raise IOError(f"User config not found: {config_file}")


def _getPluginPath(config):
    plugin_path = [USER_PLUGINS_DIR]

    if config and config.has_option("default", "plugin_path"):
        val = config.get("default", "plugin_path")
        plugin_path += [os.path.expanduser(os.path.expandvars(d)) for d
                            in val.split(':') if val]
    return plugin_path


def profileMain(args, config):  # pragma: no cover
    """This is the main function for profiling
    http://code.google.com/appengine/kb/commontasks.html#profiling
    """
    import cProfile
    import pstats

    eyed3.log.debug("driver profileMain")
    prof = cProfile.Profile()
    prof = prof.runctx("main(args)", globals(), locals())

    stream = StringIO()
    stats = pstats.Stats(prof, stream=stream)
    stats.sort_stats("time")  # Or cumulative
    stats.print_stats(100)  # 80 = how many to print

    # The rest is optional.
    stats.print_callees()
    stats.print_callers()
    sys.stderr.write("Profile data:\n%s\n" % stream.getvalue())

    return 0


def setFileScannerOpts(arg_parser, default_recursive=False, paths_metavar="PATH",
                       paths_help="Files or directory paths"):

    if default_recursive is False:
        arg_parser.add_argument("-r", "--recursive", action="store_true", dest="recursive",
                                help="Recurse into subdirectories.")
    else:
        arg_parser.add_argument("-R", "--non-recursive", action="store_true", dest="non_recursive",
                                help="Do not recurse into subdirectories.")

    arg_parser.add_argument("--exclude", action="append", metavar="PATTERN", dest="excludes",
                            help="A regular expression for path exclusion. May be specified "
                                 "multiple times.")
    arg_parser.add_argument("--fs-encoding", action="store", dest="fs_encoding",
                            default=eyed3.LOCAL_FS_ENCODING, metavar="ENCODING",
                            help="Use the specified file system encoding for filenames. "
                                 f"Default as it was detected is '{eyed3.LOCAL_FS_ENCODING}' but "
                                 "this option is still useful when reading from mounted file "
                                 "systems.")
    arg_parser.add_argument("paths", metavar=paths_metavar, nargs="*", help=paths_help)


def makeCmdLineParser(subparser=None):
    from eyed3.utils import ArgumentParser

    p = ArgumentParser(prog=eyed3.__about__.__project_name__, add_help=True)\
            if not subparser else subparser

    setFileScannerOpts(p)

    p.add_argument("-L", "--plugins", action="store_true", default=False,
                   dest="list_plugins", help="List all available plugins")
    p.add_argument("-P", "--plugin", action="store", dest="plugin",
                   default=None, metavar="NAME",
                   help=f"Specify which plugin to use. The default is '{DEFAULT_PLUGIN}'")
    p.add_argument("-C", "--config", action="store", dest="config",
                   default=None, metavar="FILE",
                   help="Supply a configuration file. The default is "
                        f"'{DEFAULT_CONFIG}', although even that is optional.")
    p.add_argument("--backup", action="store_true", dest="backup",
                   help="Plugins should honor this option such that "
                        "a backup is made of any file modified. The backup "
                        "is made in same directory with a '.orig' "
                        "extension added.")
    p.add_argument("-Q", "--quiet", action="store_true", dest="quiet",
                   default=False, help="A hint to plugins to output less.")
    p.add_argument("--no-color", action="store_true", dest="no_color",
                   help="Suppress color codes in console output. "
                        "This will happen automatically if the output is "
                        "not a TTY (e.g. when redirecting to a file)")
    p.add_argument("--no-config",
                   action="store_true", dest="no_config",
                   help=f"Do not load the default user config '{DEFAULT_CONFIG}'. "
                        "The -c/--config options are still honored if present.")

    return p


def parseCommandLine(cmd_line_args=None):

    cmd_line_args = list(cmd_line_args) if cmd_line_args else list(sys.argv[1:])

    # Remove any options not related to plugin/config for first parse. These
    # determine the parser for the next stage.
    stage_one_args = []
    idx, auto_append = 0, False
    while idx < len(cmd_line_args):
        opt = cmd_line_args[idx]
        if auto_append:
            stage_one_args.append(opt)
            auto_append = False

        if opt in ("-C", "--config", "-P", "--plugin", "--no-config"):
            stage_one_args.append(opt)
            if opt != "--no-config":
                auto_append = True
        elif (opt.startswith("-C=") or opt.startswith("--config=") or
                opt.startswith("-P=") or opt.startswith("--plugin=")):
            stage_one_args.append(opt)
        idx += 1

    parser = makeCmdLineParser()
    args = parser.parse_args(stage_one_args)

    config = _loadConfig(args)

    if args.plugin:
        # Plugin on the command line takes precedence over config.
        plugin_name = args.plugin
    elif config and config.has_option("default", "plugin"):
        # Get default plugin from config or use DEFAULT_CONFIG
        plugin_name = config.get("default", "plugin")
        if not plugin_name:
            plugin_name = DEFAULT_PLUGIN
    else:
        plugin_name = DEFAULT_PLUGIN
    assert plugin_name

    PluginClass = eyed3.plugins.load(plugin_name, paths=_getPluginPath(config))
    if PluginClass is None:
        eyed3.utils.console.printError("Plugin not found: %s" % plugin_name)
        parser.exit(1)
    plugin = PluginClass(parser)

    if config and config.has_option("default", "options"):
        cmd_line_args.extend(config.get("default", "options").split())
    if config and config.has_option(plugin_name, "options"):
        cmd_line_args.extend(config.get(plugin_name, "options").split())

    # Re-parse the command line including options from the config.
    args = parser.parse_args(args=cmd_line_args)

    args.plugin = plugin
    eyed3.log.debug("command line args: %s", args)
    eyed3.log.debug("plugin is: %s", plugin)

    return args, parser, config


def _main():
    """Entry point"""
    initLogging()

    args = None
    try:
        args, _, config = parseCommandLine()

        eyed3.utils.console.AnsiCodes.init(not args.no_color)

        mainFunc = main if args.debug_profile is False else profileMain
        retval = mainFunc(args, config)
    except KeyboardInterrupt:
        retval = 0
    except (StopIteration, IOError) as ex:
        eyed3.utils.console.printError(str(ex))
        retval = 1
    except Exception as ex:
        eyed3.utils.console.printError(f"Uncaught exception: {ex}\n")
        eyed3.log.exception(ex)
        retval = 1

        if args.debug_pdb:
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", PendingDeprecationWarning)
                    # Must delay the import of ipdb as say as possible because
                    # of https://github.com/gotcha/ipdb/issues/48
                    import ipdb as pdb  # noqa
            except ImportError:
                import pdb  # noqa

            e, m, tb = sys.exc_info()
            pdb.post_mortem(tb)

    sys.exit(retval)


if __name__ == "__main__":  # pragma: no cover
    _main()



import time
import eyed3
from eyed3.utils.console import printMsg
from eyed3.utils import formatSize, formatTime
from eyed3.id3 import versionToString
from eyed3.plugins import LoaderPlugin


class NfoPlugin(LoaderPlugin):
    NAMES = ["nfo"]
    SUMMARY = "Create NFO files for each directory scanned."
    DESCRIPTION = "Each directory scanned is treated as an album and a "\
                  "`NFO <http://en.wikipedia.org/wiki/.nfo>`_ file is "\
                  "written to standard out.\n\n"\
                  "NFO files are often found in music archives."

    def __init__(self, arg_parser):
        super(NfoPlugin, self).__init__(arg_parser)
        self.albums = {}

    def handleFile(self, f):
        super(NfoPlugin, self).handleFile(f)

        if self.audio_file and self.audio_file.tag:
            tag = self.audio_file.tag
            album = tag.album
            if album and album not in self.albums:
                self.albums[album] = []
                self.albums[album].append(self.audio_file)
            elif album:
                self.albums[album].append(self.audio_file)

    def handleDone(self):
        if not self.albums:
            printMsg("No albums found.")
            return

        for album in self.albums:
            audio_files = self.albums[album]
            if not audio_files:
                continue
            audio_files.sort(key=lambda af: (af.tag.track_num[0] or 999,
                                             af.tag.track_num[1] or 999))

            max_title_len = 0
            avg_bitrate = 0
            encoder_info = ''
            for audio_file in audio_files:
                tag = audio_file.tag
                # Compute maximum title length
                title_len = len(tag.title or "")
                if title_len > max_title_len:
                    max_title_len = title_len
                # Compute average bitrate
                avg_bitrate += audio_file.info.bit_rate[1]
                # Grab the last lame version in case not all files have one
                if "encoder_version" in audio_file.info.lame_tag:
                    version = audio_file.info.lame_tag['encoder_version']
                    encoder_info = (version or encoder_info)
            avg_bitrate = avg_bitrate / len(audio_files)

            printMsg("")
            printMsg("Artist   : %s" % audio_files[0].tag.artist)
            printMsg("Album    : %s" % album)
            printMsg("Released : %s" %
                     (audio_files[0].tag.original_release_date or
                      audio_files[0].tag.release_date))
            printMsg("Recorded : %s" % audio_files[0].tag.recording_date)
            genre = audio_files[0].tag.genre
            if genre:
                genre = genre.name
            else:
                genre = ""
            printMsg("Genre    : %s" % genre)

            printMsg("")
            printMsg("Source  : ")
            printMsg("Encoder : %s" % encoder_info)
            printMsg("Codec   : mp3")
            printMsg("Bitrate : ~%s K/s @ %s Hz, %s" %
                     (avg_bitrate, audio_files[0].info.sample_freq,
                      audio_files[0].info.mode))
            printMsg("Tag     : ID3 %s" %
                     versionToString(audio_files[0].tag.version))

            printMsg("")
            printMsg("Ripped By: ")

            printMsg("")
            printMsg("Track Listing")
            printMsg("-------------")
            count = 0
            total_time = 0
            total_size = 0
            for audio_file in audio_files:
                tag = audio_file.tag
                count += 1

                title = tag.title or ""
                title_len = len(title)
                padding = " " * ((max_title_len - title_len) + 3)
                time_secs = audio_file.info.time_secs
                total_time += time_secs
                total_size += audio_file.info.size_bytes

                zero_pad = "0" * (len(str(len(audio_files))) - len(str(count)))
                printMsg(" %s%d. %s%s(%s)" %
                         (zero_pad, count, title, padding,
                          formatTime(time_secs)))

            printMsg("")
            printMsg("Total play time : %s" %
                     formatTime(total_time))
            printMsg("Total size      : %s" %
                     formatSize(total_size))

            printMsg("")
            printMsg("=" * 78)
            printMsg(".NFO file created with eyeD3 %s on %s" %
                     (eyed3.version, time.asctime()))
            printMsg("For more information about eyeD3 go to %s" %
                     "http://eyeD3.nicfit.net/")
            printMsg("=" * 78)
